import sys

from typeguard import typechecked

"""Entrance to invoke any backward compatibility patch

This module is for developers to quickly locate backward compatibility pathes
in Kestrel code and remove them through time.
"""


@typechecked
def is_python_older_than_minor_version(minor: int) -> bool:
    return sys.version_info.minor < minor



import pytest
from pandas import DataFrame

from kestrel.exceptions import InvalidDataSource, InvalidSeralizedInstruction
from kestrel.ir.instructions import (CACHE_INTERFACE_IDENTIFIER, Construct,
                                     DataSource, Variable,
                                     get_instruction_class,
                                     instruction_from_dict,
                                     instruction_from_json)


def test_instruction_post_init():
    v = Variable("asdf", "foo", "bar")
    j = v.to_dict()
    assert "id" in j
    assert "instruction" in j
    assert j["instruction"] == "Variable"


def test_stable_id():
    v = Variable("asdf", "foo", "bar")
    _id = v.id
    v.name = "qwer"
    assert v.id == _id


def test_stable_hash():
    s = DataSource("stixshifter://abc")
    h1 = hash(s)
    s.datasource = "abcd"
    h2 = hash(s)
    assert h1 == h2


def test_eq():
    s1 = DataSource("stixshifter://abc")
    s2 = DataSource("stixshifter://abc")
    s3 = instruction_from_dict(s1.to_dict())
    assert s1 != s2
    assert s1 == s3


def test_get_instruction_class():
    cls = get_instruction_class("Variable")
    v = cls("asdf", "foo", "bar")
    assert cls == Variable
    assert isinstance(v, Variable)


def test_add_source():
    s = DataSource("stixshifter://abc")
    j = s.to_dict()
    assert j["interface"] == "stixshifter"
    assert j["datasource"] == "abc"
    assert "id" in j
    assert "instruction" in j
    assert "uri" not in j
    assert "default_interface" not in j

    x = DataSource("abc", "stixshifter")
    assert x.interface == "stixshifter"
    assert x.datasource == "abc"

    with pytest.raises(InvalidDataSource):
        DataSource("sss://eee://ccc")

    with pytest.raises(InvalidDataSource):
        DataSource("sss")


def test_construct():
    data = [ {"name": "cmd.exe", "pid": 123}
           , {"name": "explorer.exe", "pid": 99}
           , {"name": "firefox.exe", "pid": 201}
           , {"name": "chrome.exe", "pid": 205}
           ]
    c = Construct(data)
    assert c.data.equals(DataFrame(data))
    assert c.interface == CACHE_INTERFACE_IDENTIFIER


def test_instruction_from_dict():
    v = Variable("asdf", "foo", "bar")
    d = v.to_dict()
    w = instruction_from_dict(d)
    assert w == v

    del d["id"]
    with pytest.raises(InvalidSeralizedInstruction):
        instruction_from_dict(d)


def test_instruction_from_json():
    v = Variable("asdf", "foo", "bar")
    j = v.to_json()
    w = instruction_from_json(j)
    assert w == v



################################################################
#
# Kestrel Command-line Utilities
# - kestrel
# - ikestrel
#
################################################################

import argparse
import cmd
import logging

from kestrel.exceptions import KestrelError
from kestrel.session import Session
from pandas import DataFrame


def add_logging_handler(handler, if_debug):
    fmt = "%(asctime)s %(levelname)s %(name)s %(message)s"
    datefmt = "%H:%M:%S"
    formatter = logging.Formatter(fmt=fmt, datefmt=datefmt)

    handler.setFormatter(formatter)

    root_logger = logging.getLogger()
    current_logging_level = root_logger.getEffectiveLevel()
    root_logger.addHandler(handler)
    root_logger.setLevel(logging.DEBUG if if_debug else logging.INFO)

    return handler, current_logging_level


def kestrel():
    parser = argparse.ArgumentParser(description="Kestrel Interpreter")
    parser.add_argument("huntflow", help="huntflow in .hf file")
    parser.add_argument(
        "-v", "--verbose", help="print verbose log", action="store_true"
    )
    parser.add_argument(
        "--debug", help="debug level log (default is info level)", action="store_true"
    )
    args = parser.parse_args()

    if args.verbose:
        add_logging_handler(logging.StreamHandler(), args.debug)

    with Session() as session:
        with open(args.huntflow, "r") as fp:
            huntflow = fp.read()
        outputs = session.execute(huntflow)
        results = "\n\n".join(
            [o.to_string() if isinstance(o, DataFrame) else str(o) for o in outputs]
        )
        print(results)


# TODO: fix #405 so we do not need this
CMDS = [  # command_no_result from kestrel.lark
    "APPLY",
    "DISP",
    "INFO",
    "SAVE",
]


def display_outputs(outputs):
    for i in outputs:
        print(i)


class IKestrel(cmd.Cmd):
    prompt = "> "

    def __init__(self, session: Session):
        self.session = session
        self.buf = ""
        super().__init__()

    def default(self, line: str):
        try:
            outputs = self.session.execute(line)
            display_outputs(outputs)
        except KestrelError as e:
            print(e)

    def completenames(self, text, *ignored):
        code, _start, _end = ignored
        if code.isupper():
            # Probably a command?
            results = [i for i in CMDS if i.startswith(code)]
        else:
            # Try all commands and vars
            results = [i for i in CMDS if i.lower().startswith(code)]
            results += [
                i for i in self.session.get_variable_names() if i.startswith(code)
            ]
        return results

    def completedefault(self, *ignored):
        _, code, start, end = ignored
        results = self.session.do_complete(code, end)
        stub = code[start:]
        return [stub + suffix for suffix in results]

    def do_EOF(self, _line: str):
        print()
        return True


def ikestrel():
    parser = argparse.ArgumentParser(description="Kestrel Interpreter")
    parser.add_argument(
        "-v", "--verbose", help="print verbose log", action="store_true"
    )
    parser.add_argument(
        "--debug", help="debug level log (default is info level)", action="store_true"
    )
    args = parser.parse_args()

    if args.verbose:
        add_logging_handler(logging.StreamHandler(), args.debug)

    with Session() as s:
        ik = IKestrel(s)
        ik.cmdloop()



import logging
import os
from copy import copy
from tempfile import mkstemp
from typing import Any, Iterable, Mapping, MutableMapping, Optional
from uuid import UUID

import sqlalchemy
from dateutil.parser import parse as dt_parser
from kestrel.cache.base import AbstractCache
from kestrel.config.internal import VIRTUAL_CACHE_VAR_DATA
from kestrel.display import GraphletExplanation, NativeQuery
from kestrel.interface.codegen.sql import SqlTranslator
from kestrel.interface.codegen.utils import variable_attributes_to_dataframe
from kestrel.ir.graph import IRGraphEvaluable
from kestrel.ir.instructions import (
    Construct,
    Explain,
    Filter,
    Information,
    Instruction,
    Return,
    SolePredecessorTransformingInstruction,
    SourceInstruction,
    TransformingInstruction,
    Variable,
)
from pandas import DataFrame, read_sql
from typeguard import typechecked

_logger = logging.getLogger(__name__)


class SqlCacheTranslator(SqlTranslator):
    def __init__(self, from_obj, from_obj_schema=None, projection_base_field=None):
        super().__init__(
            sqlalchemy.dialects.sqlite.dialect(),
            from_obj,
            from_obj_schema,
            projection_base_field,
            None,
            dt_parser,
            "time",
        )  # FIXME: need mapping for timestamp?


@typechecked
class SqlCache(AbstractCache):
    def __init__(
        self,
        initial_cache: Optional[Mapping[UUID, DataFrame]] = None,
        debug: bool = False,
    ):
        super().__init__()

        if debug:
            self.db_path = "local.db"
        else:
            _, self.db_path = mkstemp(suffix=".db")

        # for an absolute file path, the three slashes are followed by the absolute path
        # for a relative path, it's also three slashes?
        self.engine = sqlalchemy.create_engine(f"sqlite:///{self.db_path}")
        self.connection = self.engine.connect()

        # besides self.cache_catalog, which stores instruction.id to table name mapping
        # we also stores instruction.id to table schema mapping for ProjectEntity use
        self.cache_catalog_schemas = {}

        if initial_cache:
            for instruction_id, data in initial_cache.items():
                self[instruction_id] = data

    def __del__(self):
        self.connection.close()
        os.remove(self.db_path)

    def __getitem__(self, instruction_id: UUID) -> DataFrame:
        return read_sql(self.cache_catalog[instruction_id], self.connection)

    def __delitem__(self, instruction_id: UUID):
        table_name = self.cache_catalog[instruction_id]
        self.connection.execute(sqlalchemy.text(f'DROP TABLE "{table_name}"'))
        del self.cache_catalog[instruction_id]

    def __setitem__(
        self,
        instruction_id: UUID,
        data: DataFrame,
    ):
        table_name = instruction_id.hex
        if table_name not in self.cache_catalog:
            self.cache_catalog[instruction_id] = table_name
            data.to_sql(
                table_name, con=self.connection, if_exists="replace", index=False
            )
            self.cache_catalog_schemas[instruction_id] = list(data)
        else:
            _logger.debug(f"instruction already cached: {instruction_id}, {data}")

    def get_virtual_copy(self) -> AbstractCache:
        v = copy(self)
        v.cache_catalog = copy(self.cache_catalog)
        v.__class__ = SqlCacheVirtual
        return v

    def evaluate_graph(
        self,
        graph: IRGraphEvaluable,
        cache: MutableMapping[UUID, Any],
        instructions_to_evaluate: Optional[Iterable[Instruction]] = None,
    ) -> Mapping[UUID, DataFrame]:
        mapping = {}
        if not instructions_to_evaluate:
            instructions_to_evaluate = graph.get_sink_nodes()
        for instruction in instructions_to_evaluate:
            _logger.debug(f"evaluate instruction: {instruction}")
            translator = self._evaluate_instruction_in_graph(graph, instruction)
            # TODO: may catch error in case evaluation starts from incomplete SQL
            _logger.debug(f"SQL query generated: {translator.result_w_literal_binds()}")
            df = read_sql(translator.result(), self.connection)

            # handle Information command
            if isinstance(instruction, Return):
                trunk, _ = graph.get_trunk_n_branches(instruction)
                if isinstance(trunk, Information):
                    df = variable_attributes_to_dataframe(df)

            mapping[instruction.id] = df
        return mapping

    def explain_graph(
        self,
        graph: IRGraphEvaluable,
        cache: MutableMapping[UUID, Any],
        instructions_to_explain: Optional[Iterable[Instruction]] = None,
    ) -> Mapping[UUID, GraphletExplanation]:
        mapping = {}
        if not instructions_to_explain:
            instructions_to_explain = graph.get_sink_nodes()
        for instruction in instructions_to_explain:
            dep_graph = graph.duplicate_dependent_subgraph_of_node(instruction)
            graph_dict = dep_graph.to_dict()
            translator = self._evaluate_instruction_in_graph(graph, instruction)
            query = NativeQuery("SQL", str(translator.result_w_literal_binds()))
            mapping[instruction.id] = GraphletExplanation(graph_dict, query)
        return mapping

    def _evaluate_instruction_in_graph(
        self,
        graph: IRGraphEvaluable,
        instruction: Instruction,
        graph_genuine_copy: Optional[IRGraphEvaluable] = None,
        subquery_memory: Optional[Mapping[UUID, SqlCacheTranslator]] = None,
    ) -> SqlCacheTranslator:
        """Evaluate the instruction in the graph

        This method recursively traverse the graph from the instruction node to
        evaluate the instruction with all its dependencies.

        To avoid repeated traversal/evaluation of the same subgraph/subtree,
        for each Variable instruction/node, the method performs dynamic
        programming in the form of memorization subgraph results as CTEs. This
        advanced feature requires the underlying SQL engine to support common
        table expression (CTE), which may not be possible for query engines
        like SQL on OpenSearch (Kestrel OpenSearch interface uses embedded
        subquery instead of CTE).

        To avoid unexpected Python behavior
        https://docs.quantifiedcode.com/python-anti-patterns/correctness/mutable_default_value_as_argument.html
        We use `None` as default value instead of `{}`

        Parameters:
            graph: the graph to traverse, node of which will be modified during evaluation
            instruction: the instruction to evaluate/return
            graph_genuine_copy: the original graph, deep copy, no modification, for traversal use
            subquery_memory: memorize the subgraph traversed/evaluated

        Returns:
            A translator (SQL statements) to be executed
        """
        if graph_genuine_copy is None:
            graph_genuine_copy = graph.deepcopy()

        if subquery_memory is None:
            subquery_memory = {}

        if instruction.id in self:
            # cached in sqlite
            table_name = self.cache_catalog[instruction.id]
            source_schema = self.cache_catalog_schemas[instruction.id]
            translator = SqlCacheTranslator(table_name, source_schema)

        elif isinstance(instruction, SourceInstruction):
            if isinstance(instruction, Construct):
                # cache the data
                self[instruction.id] = DataFrame(instruction.data)
                # pull the data to start a SqlCacheTranslator
                table_name = self.cache_catalog[instruction.id]
                source_schema = self.cache_catalog_schemas[instruction.id]
                translator = SqlCacheTranslator(table_name, source_schema)
            else:
                raise NotImplementedError(f"Unknown instruction type: {instruction}")

        elif isinstance(instruction, TransformingInstruction):
            if instruction.id in subquery_memory:
                # this is a Variable, already evaluated
                # just create a new use/translator from this Variable
                translator = subquery_memory[instruction.id]
            else:
                trunk, r2n = graph.get_trunk_n_branches(instruction)
                translator = self._evaluate_instruction_in_graph(
                    graph, trunk, graph_genuine_copy, subquery_memory
                )

                if isinstance(instruction, SolePredecessorTransformingInstruction):
                    if isinstance(instruction, (Return, Explain)):
                        pass
                    elif isinstance(instruction, Variable):
                        subquery_memory[instruction.id] = translator
                        translator = SqlCacheTranslator(
                            translator.query.cte(name=instruction.name),
                            translator.projected_schema,
                            translator.projection_base_field,
                        )
                    else:
                        translator.add_instruction(instruction)

                elif isinstance(instruction, Filter):
                    # replace each ReferenceValue with a subquery
                    # note that this subquery will be used as a value for the .in_ operator
                    # we should not use .subquery() here but just `Select` class
                    # otherwise, will get warning:
                    #   SAWarning: Coercing Subquery object into a select() for use in IN();
                    #   please pass a select() construct explicitly
                    instruction.resolve_references(
                        lambda x: self._evaluate_instruction_in_graph(
                            graph, r2n[x], graph_genuine_copy, subquery_memory
                        ).query
                    )
                    translator.add_instruction(instruction)

                else:
                    raise NotImplementedError(
                        f"Unknown instruction type: {instruction}"
                    )

        else:
            raise NotImplementedError(f"Unknown instruction type: {instruction}")

        return translator


@typechecked
class SqlCacheVirtual(SqlCache):
    def __getitem__(self, instruction_id: UUID) -> DataFrame:
        if instruction_id in self.cache_catalog:
            try:
                df = read_sql(self.cache_catalog[instruction_id], self.connection)
            except:
                df = VIRTUAL_CACHE_VAR_DATA
        else:
            raise KeyError(instruction_id)
        return df

    def __delitem__(self, instruction_id: UUID):
        del self.cache_catalog[instruction_id]

    def __setitem__(self, instruction_id: UUID, data: DataFrame):
        self.cache_catalog[instruction_id] = instruction_id.hex + "v"
        self.cache_catalog_schemas[instruction_id] = ["*"]

    def __del__(self):
        pass



import json
import os
from uuid import uuid4

import pytest
from pandas import DataFrame, read_csv

from kestrel import Session
from kestrel.cache import SqlCache
from kestrel.config.internal import CACHE_INTERFACE_IDENTIFIER
from kestrel.display import GraphExplanation
from kestrel.frontend.parser import parse_kestrel_and_update_irgraph
from kestrel.ir.graph import IRGraph, IRGraphEvaluable
from kestrel.ir.instructions import Construct, SerializableDataFrame
from kestrel.exceptions import EntityNotFound


@pytest.fixture
def process_creation_events():
    # return a two-node graph:
    #   - Construct: table from logs_ocsf_process_creation.csv
    #   - Variable es: events pointing to the Construct
    graph = IRGraph()
    parse_kestrel_and_update_irgraph("es = NEW event [ {'id': 1} ]", graph, {})
    data_node = graph.get_nodes_by_type(Construct)[0]
    test_dir = os.path.dirname(os.path.abspath(__file__))
    data_node.data = SerializableDataFrame(read_csv(os.path.join(test_dir, "logs_ocsf_process_creation.csv")))
    return graph


def test_execute_in_cache():
    hf = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
browsers = proclist WHERE name != "cmd.exe"
DISP browsers
cmd = proclist WHERE name = "cmd.exe"
DISP cmd ATTR pid
"""
    b1 = DataFrame([ {"name": "explorer.exe", "pid": 99}
                   , {"name": "firefox.exe", "pid": 201}
                   , {"name": "chrome.exe", "pid": 205}
                   ])
    b2 = DataFrame([ {"pid": 123} ])
    with Session() as session:
        res = session.execute_to_generate(hf)
        assert b1.equals(next(res))
        assert b2.equals(next(res))
        with pytest.raises(StopIteration):
            next(res)


def test_execute_in_cache_stix_process():
    hf = """
proclist = NEW process [ {"file.name": "cmd.exe", "pid": 123}
                       , {"file.name": "explorer.exe", "pid": 99}
                       , {"file.name": "firefox.exe", "pid": 201}
                       , {"file.name": "chrome.exe", "pid": 205}
                       ]
DISP proclist ATTR binary_ref.name
"""
    b1 = DataFrame([ {"file.name": "cmd.exe"}
                   , {"file.name": "explorer.exe"}
                   , {"file.name": "firefox.exe"}
                   , {"file.name": "chrome.exe"}
                   ])
    with Session() as session:
        res = session.execute_to_generate(hf)
        assert b1.equals(next(res))
        with pytest.raises(StopIteration):
            next(res)


@pytest.mark.skip("TODO: need attr mapping for Construct")
def test_execute_in_cache_stix_process_ocsf_disp_attr():
    hf = """
proclist = NEW process [ {"binary_ref.name": "cmd.exe", "pid": 123}
                       , {"binary_ref.name": "explorer.exe", "pid": 99}
                       , {"binary_ref.name": "firefox.exe", "pid": 201}
                       , {"binary_ref.name": "chrome.exe", "pid": 205}
                       ]
DISP proclist ATTR file.name
"""
    b1 = DataFrame([ {"file.name": "cmd.exe"}
                   , {"file.name": "explorer.exe"}
                   , {"file.name": "firefox.exe"}
                   , {"file.name": "chrome.exe"}
                   ])
    with Session() as session:
        res = session.execute_to_generate(hf)
        assert b1.equals(next(res))
        with pytest.raises(StopIteration):
            next(res)


def test_execute_in_cache_stix_process_filtered():
    hf = """
proclist = NEW process [ {"file.name": "cmd.exe", "pid": 123}
                       , {"file.name": "explorer.exe", "pid": 99}
                       , {"file.name": "firefox.exe", "pid": 201}
                       , {"file.name": "chrome.exe", "pid": 205}
                       ]
browsers = proclist WHERE binary_ref.name in ('chrome.exe', 'firefox.exe')
DISP browsers ATTR binary_ref.name, pid
"""
    b1 = DataFrame([ {"file.name": "firefox.exe", "pid": 201}
                   , {"file.name": "chrome.exe", "pid": 205}
                   ])
    with Session() as session:
        res = session.execute_to_generate(hf)
        df = next(res)
        assert b1.equals(df)
        with pytest.raises(StopIteration):
            next(res)


def test_execute_in_cache_stix_process_with_ref_and_multi_returns():
    hf = """
proclist = NEW process [ {"file.name": "cmd.exe", "pid": 123}
                       , {"file.name": "explorer.exe", "pid": 99}
                       , {"file.name": "firefox.exe", "pid": 201}
                       , {"file.name": "chrome.exe", "pid": 205}
                       ]
newvar = proclist WHERE binary_ref.name = "cmd.exe"
DISP proclist ATTR binary_ref.name
newvar2 = proclist WHERE binary_ref.name IN ("explorer.exe", "cmd.exe")
newvar3 = newvar2 WHERE pid IN newvar.pid
DISP newvar3 ATTR binary_ref.name
"""
    b1 = DataFrame([ {"file.name": "cmd.exe"}
                   , {"file.name": "explorer.exe"}
                   , {"file.name": "firefox.exe"}
                   , {"file.name": "chrome.exe"}
                   ])
    b2 = DataFrame([ {"file.name": "cmd.exe"}
                   ])
    with Session() as session:
        res = session.execute_to_generate(hf)
        assert b1.equals(next(res))
        assert b2.equals(next(res))
        with pytest.raises(StopIteration):
            next(res)


def test_execute_in_cache_stix_file():
    data = [ {"name": "cmd.exe", "hashes.MD5": "AD7B9C14083B52BC532FBA5948342B98"}
           , {"name": "powershell.exe", "hashes.MD5": "04029E121A0CFA5991749937DD22A1D9"}
    ]
    hf = f"""
filelist = NEW file {json.dumps(data)}
DISP filelist ATTR name, hashes.MD5
"""
    b1 = DataFrame(data)
    with Session() as session:
        res = session.execute_to_generate(hf)
        assert b1.equals(next(res))
        with pytest.raises(StopIteration):
            next(res)


def test_double_deref_in_cache():
    # When the Filter node is dereferred twice
    # The node should be deepcopied each time to avoid issue
    hf = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
px = proclist WHERE name != "cmd.exe" AND pid = 205
chrome = proclist WHERE pid IN px.pid
DISP chrome
DISP chrome
"""
    df = DataFrame([ {"name": "chrome.exe", "pid": 205} ])
    with Session() as session:
        res = session.execute_to_generate(hf)
        assert df.equals(next(res))
        assert df.equals(next(res))
        with pytest.raises(StopIteration):
            next(res)


def test_explain_in_cache():
    hf = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
browsers = proclist WHERE name != "cmd.exe"
chrome = browsers WHERE pid = 205
EXPLAIN chrome
"""
    with Session() as session:
        ress = session.execute_to_generate(hf)
        res = next(ress)
        assert isinstance(res, GraphExplanation)
        assert len(res.graphlets) == 1
        ge = res.graphlets[0]
        assert ge.graph == IRGraphEvaluable(session.irgraph).to_dict()
        construct = session.irgraph.get_nodes_by_type(Construct)[0]
        assert ge.action.language == "SQL"
        stmt = ge.action.statement.replace('"', '')
        assert stmt == f"WITH proclist AS \n(SELECT DISTINCT * \nFROM {construct.id.hex}v), \nbrowsers AS \n(SELECT DISTINCT * \nFROM proclist \nWHERE name != 'cmd.exe'), \nchrome AS \n(SELECT DISTINCT * \nFROM browsers \nWHERE pid = 205)\n SELECT DISTINCT * \nFROM chrome"
        with pytest.raises(StopIteration):
            next(ress)


def test_multi_interface_explain():

    class DataLake(SqlCache):
        @staticmethod
        def schemes():
            return ["datalake"]

    class Gateway(SqlCache):
        @staticmethod
        def schemes():
            return ["gateway"]

    with Session() as session:
        stmt1 = """
procs = NEW process [ {"name": "cmd.exe", "pid": 123}
                    , {"name": "explorer.exe", "pid": 99}
                    , {"name": "firefox.exe", "pid": 201}
                    , {"name": "chrome.exe", "pid": 205}
                    ]
DISP procs
"""
        session.execute(stmt1)
        session.interface_manager[CACHE_INTERFACE_IDENTIFIER].__class__ = DataLake
        session.irgraph.get_nodes_by_type_and_attributes(Construct, {"interface": CACHE_INTERFACE_IDENTIFIER})[0].interface = "datalake"

        new_cache = SqlCache()
        session.interface_manager.interfaces.append(new_cache)
        stmt2 = """
nt = NEW event [ {"abc": 123, "source": "192.168.1.1", "destination": "1.1.1.1"}
               , {"abc": 205, "source": "192.168.1.1", "destination": "1.1.1.2"}
               ]
DISP nt
"""
        session.execute(stmt2)
        session.interface_manager[CACHE_INTERFACE_IDENTIFIER].__class__ = Gateway
        session.irgraph.get_nodes_by_type_and_attributes(Construct, {"interface": CACHE_INTERFACE_IDENTIFIER})[0].interface = "gateway"

        new_cache = SqlCache()
        session.interface_manager.interfaces.append(new_cache)
        stmt3 = """
domain = NEW domain [ {"ip": "1.1.1.1", "domain": "cloudflare.com"}
                    , {"ip": "1.1.1.2", "domain": "xyz.cloudflare.com"}
                    ]
DISP domain
"""
        session.execute(stmt3)

        stmt = """
p2 = procs WHERE name IN ("firefox.exe", "chrome.exe")
ntx = nt WHERE abc IN p2.pid
d2 = domain WHERE ip IN ntx.destination
EXPLAIN d2
DISP d2
"""
        ress = session.execute_to_generate(stmt)
        disp = next(ress)
        df_res = next(ress)

        with pytest.raises(StopIteration):
            next(ress)

        assert isinstance(disp, GraphExplanation)
        assert len(disp.graphlets) == 4

        # DISP procs
        assert len(disp.graphlets[0].graph["nodes"]) == 5
        query = disp.graphlets[0].action.statement.replace('"', '')
        procs = session.irgraph.get_variable("procs")
        c1 = next(session.irgraph.predecessors(procs))
        assert query == f"WITH procs AS \n(SELECT DISTINCT * \nFROM {c1.id.hex}), \np2 AS \n(SELECT DISTINCT * \nFROM procs \nWHERE name IN ('firefox.exe', 'chrome.exe'))\n SELECT DISTINCT pid \nFROM p2"

        # DISP nt
        assert len(disp.graphlets[1].graph["nodes"]) == 2
        query = disp.graphlets[1].action.statement.replace('"', '')
        nt = session.irgraph.get_variable("nt")
        c2 = next(session.irgraph.predecessors(nt))
        assert query == f"WITH nt AS \n(SELECT DISTINCT * \nFROM {c2.id.hex})\n SELECT DISTINCT * \nFROM nt"

        # DISP domain
        assert len(disp.graphlets[2].graph["nodes"]) == 2
        query = disp.graphlets[2].action.statement.replace('"', '')
        domain = session.irgraph.get_variable("domain")
        c3 = next(session.irgraph.predecessors(domain))
        assert query == f"WITH domain AS \n(SELECT DISTINCT * \nFROM {c3.id.hex})\n SELECT DISTINCT * \nFROM domain"

        # EXPLAIN d2
        assert len(disp.graphlets[3].graph["nodes"]) == 11
        query = disp.graphlets[3].action.statement.replace('"', '')
        p2 = session.irgraph.get_variable("p2")
        p2pa = next(session.irgraph.successors(p2))
        assert query == f"WITH ntx AS \n(SELECT DISTINCT * \nFROM {nt.id.hex}v \nWHERE abc IN (SELECT DISTINCT * \nFROM {p2pa.id.hex}v)), \nd2 AS \n(SELECT DISTINCT * \nFROM {domain.id.hex}v \nWHERE ip IN (SELECT DISTINCT destination \nFROM ntx))\n SELECT DISTINCT * \nFROM d2"

        df_ref = DataFrame([{"ip": "1.1.1.2", "domain": "xyz.cloudflare.com"}])
        assert df_ref.equals(df_res)


def test_apply_on_construct():
    hf = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
APPLY python://something ON proclist WITH foo=abc,bar=1,baz=1.5
DISP proclist ATTR name, foo, bar, baz
"""
    b1 = DataFrame([ {"name": "cmd.exe", "foo": "abc", "bar": 1, "baz": 1.5}
                   , {"name": "explorer.exe", "foo": "abc", "bar": 1, "baz": 1.5}
                   , {"name": "firefox.exe", "foo": "abc", "bar": 1, "baz": 1.5}
                   , {"name": "chrome.exe", "foo": "abc", "bar": 1, "baz": 1.5}
                   ])
    with Session() as session:
        # Add test analytic
        test_dir = os.path.dirname(os.path.abspath(__file__))
        session.interface_manager["python"].config["something"] = {
            "module": os.path.join(test_dir, "test_analytic.py"),
            "func": "do_something"
        }
        res = session.execute_to_generate(hf)
        disp = next(res)
        assert b1.equals(disp)
        with pytest.raises(StopIteration):
            next(res)


def test_apply_on_construct_use_env():
    hf = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
APPLY python://something ON proclist WITH name=foo,value=1
DISP proclist ATTR name, foo
"""
    b1 = DataFrame([ {"name": "cmd.exe", "foo": 1}
                   , {"name": "explorer.exe", "foo": 1}
                   , {"name": "firefox.exe", "foo": 1}
                   , {"name": "chrome.exe", "foo": 1}
                   ])
    with Session() as session:
        # Add test analytic
        test_dir = os.path.dirname(os.path.abspath(__file__))
        session.interface_manager["python"].config["something"] = {
            "module": os.path.join(test_dir, "test_analytic.py"),
            "func": "do_something_env"
        }
        res = session.execute_to_generate(hf)
        disp = next(res)
        assert b1.equals(disp)
        with pytest.raises(StopIteration):
            next(res)


def test_explain_find_event_to_entity(process_creation_events):
    with Session() as session:
        session.irgraph = process_creation_events
        res = session.execute("procs = FIND process RESPONDED es WHERE device.os = 'Linux' EXPLAIN procs")[0]
        construct = session.irgraph.get_nodes_by_type(Construct)[0]
        stmt = res.graphlets[0].action.statement.replace('"', '')
        # cache.sql will use "*" as columns for __setitem__ in virtual cache
        # so the result is different from test_cache_sqlite::test_explain_find_event_to_entity
        assert stmt == f"WITH es AS \n(SELECT DISTINCT * \nFROM {construct.id.hex}v), \nprocs AS \n(SELECT DISTINCT * \nFROM es \nWHERE device.os = \'Linux\')\n SELECT DISTINCT * \nFROM procs"


def test_get_nonexist_entity(process_creation_events):
    with Session() as session:
        session.irgraph = process_creation_events
        with pytest.raises(EntityNotFound):
            session.execute("reg = FIND reg_key RESPONDED es WHERE device.os = 'Linux' DISP reg")



from __future__ import annotations

import importlib
import inspect
import itertools
import logging
import pkgutil
import sys
from copy import copy
from typing import Iterable, Mapping, Type

from kestrel.config.internal import CACHE_INTERFACE_IDENTIFIER
from kestrel.exceptions import (
    ConflictingInterfaceScheme,
    InterfaceNotConfigured,
    InterfaceNotFound,
    InvalidInterfaceImplementation,
)
from kestrel.interface.base import MODULE_PREFIX, AbstractInterface
from typeguard import typechecked

_logger = logging.getLogger(__name__)


# basically a scheme to interface mapping
@typechecked
class InterfaceManager(Mapping):
    def __init__(self, init_interfaces: Iterable[AbstractInterface] = []):
        interface_classes = _load_interface_classes()
        self.interfaces = list(init_interfaces)  # copy/recreate the list
        for iface_cls in interface_classes:
            try:
                iface = iface_cls()
                _logger.debug(f"Initialize interface {iface_cls.__name__}")
                self.interfaces.append(iface)
            except InterfaceNotConfigured as e:
                _logger.debug(f"Interface {iface_cls.__name__} not configured; ignored")

    def __getitem__(self, scheme: str) -> AbstractInterface:
        for interface in self.interfaces:
            if scheme in interface.schemes():
                return interface
        else:
            raise InterfaceNotFound(f"no interface loaded for scheme {scheme}")

    def __iter__(self) -> Iterable[str]:
        return filter(
            lambda x: x != CACHE_INTERFACE_IDENTIFIER,
            itertools.chain(*[i.schemes() for i in self.interfaces]),
        )

    def __len__(self) -> int:
        return sum(1 for _ in iter(self))

    def copy_with_virtual_cache(self) -> InterfaceManager:
        im = copy(self)
        # shallow copy refers to the same list, so create/copy a new one
        im.interfaces = copy(im.interfaces)
        # now swap in virtual cache
        cache = im[CACHE_INTERFACE_IDENTIFIER]
        im.interfaces.remove(cache)
        im.interfaces.append(cache.get_virtual_copy())
        return im

    def del_cache(self):
        cache = self[CACHE_INTERFACE_IDENTIFIER]
        self.interfaces.remove(cache)
        del cache

    def schemes(self, interface_type: type) -> Iterable[str]:
        return filter(
            lambda x: x != CACHE_INTERFACE_IDENTIFIER,
            itertools.chain(
                *[i.schemes() for i in self.interfaces if isinstance(i, interface_type)]
            ),
        )

    def list_datasources_from_scheme(self, scheme: str) -> Iterable[str]:
        return self[scheme].get_datasources()


def _load_interface_classes():
    interface_clss = []
    for itf_pkg_name in _list_interface_pkg_names():
        mod = importlib.import_module(itf_pkg_name)
        _logger.debug(f"Imported {mod} from package {itf_pkg_name}")
        cls = inspect.getmembers(
            sys.modules[itf_pkg_name], _is_class(AbstractInterface)
        )
        if not cls:
            raise InvalidInterfaceImplementation(
                f'no interface class found in package "{itf_pkg_name}"'
            )
        elif len(cls) > 1:
            raise InvalidInterfaceImplementation(
                f'more than one interface class found in package "{itf_pkg_name}"'
            )
        else:
            interface_cls = cls[0][1]
            _guard_scheme_conflict(interface_cls, interface_clss)
            interface_clss.append(interface_cls)
    return interface_clss


def _list_interface_pkg_names():
    pkg_names = [x.name for x in pkgutil.iter_modules()]
    itf_names = [pkg for pkg in pkg_names if pkg.startswith(MODULE_PREFIX)]
    return itf_names


def _is_class(cls):
    return lambda obj: inspect.isclass(obj) and issubclass(obj, cls)


@typechecked
def _guard_scheme_conflict(
    new_interface: Type[AbstractInterface],
    interfaces: Iterable[Type[AbstractInterface]],
):
    for interface in interfaces:
        for scheme_new in new_interface.schemes():
            for scheme_old in interface.schemes():
                if scheme_new == scheme_old:
                    raise ConflictingInterfaceScheme(
                        f"scheme: {scheme_new} conflicting between {new_interface.__name__} and {interface.__name__}"
                    )



import logging
from contextlib import AbstractContextManager
from os import environ
from typing import Iterable
from uuid import uuid4

from kestrel.analytics import PythonAnalyticsInterface
from kestrel.cache import SqlCache
from kestrel.config import load_kestrel_config
from kestrel.config.internal import CACHE_INTERFACE_IDENTIFIER, VIRTUAL_CACHE_VAR_DATA
from kestrel.display import Display, GraphExplanation
from kestrel.exceptions import InstructionNotFound
from kestrel.frontend.completor import do_complete
from kestrel.frontend.parser import parse_kestrel_and_update_irgraph
from kestrel.interface import InterfaceManager
from kestrel.ir.graph import IRGraph
from kestrel.ir.instructions import DataSource, Explain, Instruction, Return
from typeguard import typechecked

_logger = logging.getLogger(__name__)


@typechecked
class Session(AbstractContextManager):
    """Kestrel huntflow execution session"""

    def __init__(self):
        self.session_id = uuid4()
        self.irgraph = IRGraph()
        self.config = load_kestrel_config()

        if "KESTREL_DEBUG" in environ:
            self.config["debug"] = True

        # load all interfaces; cache is a special interface
        cache = SqlCache(debug=self.config["debug"])

        # Python analytics are "built-in"
        pyanalytics = PythonAnalyticsInterface()

        self.interface_manager = InterfaceManager([cache, pyanalytics])

    def execute(self, huntflow_block: str) -> Iterable[Display]:
        """Execute a Kestrel huntflow block.

        Execute a Kestrel statement or multiple consecutive statements (a
        huntflow block) This method has the context of already executed
        huntflow blocks in this session, so all existing variables can be
        referred in the new huntflow block.

        Parameters:
            huntflow_block: the new huntflow block to be executed

        Returns:
            Evaluated result per Return instruction
        """
        return list(self.execute_to_generate(huntflow_block))

    def parse_and_update_graph(self, huntflow_block: str) -> Iterable[Return]:
        """Parse a Kestrel huntflow and update the IRGraph

        Parameters:
            huntflow_block: the new huntflow block to be executed

        Returns:
            Return instructions from the huntflow_block
        """
        # Transcational huntflow block parsing/updating. If failed, roll back
        # all things done for this huntflow/code block
        irgraph_snapshot = self.irgraph.copy()
        try:
            rets = parse_kestrel_and_update_irgraph(
                huntflow_block, self.irgraph, self.config["entity_identifier"]
            )
            for s in self.irgraph.get_nodes_by_type(DataSource):
                if not s.store:
                    itf = self.interface_manager[s.interface]
                    s.store = itf.get_storage_of_datasource(s.datasource)
        except Exception as e:
            self.irgraph = irgraph_snapshot
            raise e

        return rets

    def execute_to_generate(self, huntflow_block: str) -> Iterable[Display]:
        """Execute a Kestrel huntflow and put results in a generator.

        Parameters:
            huntflow_block: the new huntflow block to be executed

        Yields:
            Evaluated result per Return instruction
        """
        rets = self.parse_and_update_graph(huntflow_block)
        for ret in rets:
            yield self.evaluate_instruction(ret)

    def evaluate_instruction(self, ins: Instruction) -> Display:
        """Evaluate a single Instruction.

        Parameters:
            ins: the instruction to evaluate

        Returns:
            Evaluated result (Kestrel Display object)
        """
        if ins not in self.irgraph:
            raise InstructionNotFound(ins.to_dict())

        pred, _ = self.irgraph.get_trunk_n_branches(ins)
        is_explain = isinstance(pred, Explain)
        display = GraphExplanation([])

        _interface_manager = (
            self.interface_manager.copy_with_virtual_cache()
            if is_explain
            else self.interface_manager
        )
        _cache = _interface_manager[CACHE_INTERFACE_IDENTIFIER]

        # The current logic leads to caching results from non-cache and lastly
        # evaluate in cache.
        # TODO: may evaluate cache first, then push dependent variables to the
        # last interface to eval; this requires priority of interfaces
        while True:
            for g in self.irgraph.find_dependent_subgraphs_of_node(ins, _cache):
                interface = _interface_manager[g.interface]
                _logger.debug("eval: subgraph: %s", [i.instruction for i in g.nodes()])
                _logger.debug("eval: interface = %s", interface)
                for iid, _display in (
                    interface.explain_graph(g, _cache)
                    if is_explain
                    else interface.evaluate_graph(g, _cache)
                ).items():
                    if is_explain:
                        display.graphlets.append(_display)
                        _cache[iid] = VIRTUAL_CACHE_VAR_DATA
                    else:
                        display = _display
                        _cache[iid] = display
                    if iid == ins.id:
                        return display

    def do_complete(self, huntflow_block: str, cursor_pos: int) -> Iterable[str]:
        """Kestrel code auto-completion.

        Parameters:
            huntflow_block: Kestrel code
            cursor_pos: the position to start completion (index in ``huntflow_block``)

        Returns:
            A list of suggested strings to complete the code
        """
        return do_complete(
            huntflow_block,
            cursor_pos,
            self.interface_manager,
            [v.name for v in self.irgraph.get_variables()],
        )

    def close(self):
        """Explicitly close the session.

        This may be executed by a context manager or when the program exits.
        """
        # Note there are two conditions that trigger this function, so it is probably executed twice
        # Be careful to write the logic in this function to avoid deleting nonexist files/dirs
        if CACHE_INTERFACE_IDENTIFIER in self.interface_manager:
            self.interface_manager.del_cache()

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()



import os

from kestrel.config.utils import load_kestrel_config


def test_env_vars_in_config():

    test_config = """---
credentials:
  username: $TEST_USER
  password: $TEST_PASSWORD
    """
    os.environ["TEST_USER"] = "test-user"
    os.environ["TEST_PASSWORD"] = "test-password"
    os.environ["KESTREL_CONFIG"] = os.path.join(os.sep, "tmp", "config.yaml")

    with open(os.getenv("KESTREL_CONFIG"), "w") as fp:
        fp.write(test_config)
    config = load_kestrel_config()
    assert config["credentials"]["username"] == "test-user"
    assert config["credentials"]["password"] == "test-password"


def test_env_vars_in_config_overwrite():

    test_config = """---
credentials:
  username: ${TEST_USER}
  password: ${TEST_PASSWORD}
testattr:
  cache_directory_prefix: $KESTREL_CACHE_DIRECTORY_PREFIX
    """
    os.environ["TEST_USER"] = "test-user"
    os.environ["TEST_PASSWORD"] = "test-password"
    os.environ["KESTREL_CONFIG"] = os.path.join(os.sep, "tmp", "config.yaml")
    os.environ["KESTREL_CACHE_DIRECTORY_PREFIX"] = "Kestrel2.0-"
    with open(os.getenv("KESTREL_CONFIG"), "w") as fp:
        fp.write(test_config)
    config = load_kestrel_config()
    assert config["credentials"]["username"] == "test-user"
    assert config["credentials"]["password"] == "test-password"
    assert config["testattr"]["cache_directory_prefix"] == "Kestrel2.0-"

def test_empty_env_var_in_config():
    test_config = """---
credentials:
  username: ${TEST_USER}
  password: ${TEST_PASSWORD}
testattr:
  cache_directory_prefix: $I_DONT_EXIST
    """
    os.environ["TEST_USER"] = "test-user"
    os.environ["TEST_PASSWORD"] = "test-password"
    os.environ["KESTREL_CONFIG"] = os.path.join(os.sep, "tmp", "config.yaml")
    os.environ["KESTREL_CACHE_DIRECTORY_PREFIX"] = "Kestrel2.0-"
    with open(os.getenv("KESTREL_CONFIG"), "w") as fp:
        fp.write(test_config)
    config = load_kestrel_config()
    assert config["credentials"]["username"] == "test-user"
    assert config["credentials"]["password"] == "test-password"
    assert config["testattr"]["cache_directory_prefix"] == "$I_DONT_EXIST"

def test_yaml_load_in_config(tmp_path):
    test_config = """---
credentials:
  username: ${TEST_USER}
  password: ${TEST_PASSWORD}
loadtest:
  xyz:
    abc: abc.yaml
    """
    os.environ["TEST_USER"] = "test-user"
    os.environ["TEST_PASSWORD"] = "test-password"
    os.environ["KESTREL_CONFIG"] = os.path.join(tmp_path, "config.yaml")
    with open(os.getenv("KESTREL_CONFIG"), "w") as fp:
        fp.write(test_config)
    with open(os.path.join(tmp_path, "abc.yaml"), "w") as fp:
        fp.write("test: fake-value")
    config = load_kestrel_config()
    assert config["credentials"]["username"] == "test-user"
    assert config["credentials"]["password"] == "test-password"
    assert config["loadtest"]["xyz"]["abc"]["test"] == "fake-value"



"""Python analytics interface executes Python function as Kestrel analytics.

Use a Python Analytics
----------------------

Create a profile for each analytics in the python analytics interface config
file (YAML):

- Default path: ``~/.config/kestrel/pythonanalytics.yaml``.
- A customized path specified in the environment variable ``KESTREL_PYTHON_ANALYTICS_CONFIG``.

Example of the python analytics interface config file:

.. code-block:: yaml

    profiles:
        analytics-name-1: # the analytics name to use in the APPLY command
            module: /home/user/kestrel-analytics/analytics/piniponmap/analytics.py
            func: analytics # the analytics function in the module to call
        analytics-name-2:
            module: /home/user/kestrel-analytics/analytics/suspiciousscoring/analytics.py
            func: analytics

Develop a Python Analytics
--------------------------

A Python analytics is a python function that follows the rules:

#. The function takes in one or more Kestrel variable dumps in Pandas DataFrames.

#. The return of the function is a tuple containing either or both:

    - Updated variables. The number of variables can be either 0, e.g.,
      visualization analytics, or the same number as input Kestrel variables.
      The order of the updated variables should follow the same order as input
      variables.

    - An object to display, which can be any of the following types:

        - Kestrel display object

        - HTML element as a string

        - Matplotlib figure (by default, Pandas DataFrame plots use this)

   The display object can be either before or after updated variables. In other
   words, if the input variables are ``var1``, ``var2``, and ``var3``, the
   return of the analytics can be either of the following:

   .. code-block:: python

       # the analytics enriches variables without returning a display object
       return var1_updated, var3_updated, var3_updated

       # this is a visualization analytics and no variable updates
       return display_obj

       # the analytics does both variable updates and visualization
       return var1_updated, var3_updated, var3_updated, display_obj

       # the analytics does both variable updates and visualization
       return display_obj, var1_updated, var3_updated, var3_updated


#. Parameters in the APPLY command are passed in as keyword args when possible,
   otherwise as environment variables.  The interface will inspect the signature
   of the analytics function to determine which methods to use.  For example, the
   following analytics function would be called with keyword args taken exactly
   from the ``WITH`` part of the Kestrel statement.

   .. code-block::

   def my_analytic(df: pd.DataFrame, x: int = 0, y: float = 0.5)

   This function could be called as:

   .. code-block::

       APPLY python://my_analytic ON var1 WITH x=1, y=0.7

   The names of the environment variables are the exact parameter keys given in the
   ``APPLY`` command. For example, the following command

   .. code-block::

       APPLY python://a1 ON var1 WITH XPARAM=src_ref.value, YPARAM=number_observed

   creates environment variables ``$XPARAM`` with value ``src_ref.value`` and
   ``$YPARAM`` with value ``number_observed`` to be used by the analytics
   ``a1``. After the execution of the analytics, the environment variables will
   be roll back to the original state.

#. The Python function could spawn other processes or execute other binaries,
   where the Python function just acts like a wrapper. Check our `domain name
   lookup analytics`_ as an example.

.. _domain name lookup analytics: https://github.com/opencybersecurityalliance/kestrel-analytics/tree/release/analytics/domainnamelookup

"""

import inspect
import json
import logging
import os
import pathlib
import sys
import traceback
from contextlib import AbstractContextManager
from importlib.util import module_from_spec, spec_from_file_location
from typing import Any, Iterable, List, Mapping, MutableMapping, Optional
from uuid import UUID

from kestrel.analytics.config import get_profile, load_profiles
from kestrel.display import AnalyticOperation, GraphletExplanation
from kestrel.exceptions import (
    AnalyticsError,
    InvalidAnalytics,
    InvalidAnalyticsArgumentCount,
    InvalidAnalyticsInterfaceImplementation,
    InvalidAnalyticsOutput,
)
from kestrel.interface import AnalyticsInterface
from kestrel.ir.graph import IRGraphEvaluable
from kestrel.ir.instructions import (
    Analytic,
    Instruction,
    TransformingInstruction,
    Variable,
)
from pandas import DataFrame

_logger = logging.getLogger(__name__)


class PythonAnalyticsJob:
    """Simple config class to hold all the bits necessary to call the external analytics"""

    def __init__(
        self,
        iid: UUID,
        cache: MutableMapping[UUID, Any],
    ):
        self.cache = cache
        self.input_iid = iid
        self.output_iid: Optional[UUID] = None
        self.analytic: str = ""
        self.params: dict = {}

    def run(self, config: dict) -> DataFrame:
        module_name, func_name = get_profile(self.analytic, config)
        df = self.cache[self.input_iid]
        with PythonAnalytics(
            self.analytic, module_name, func_name, self.params
        ) as func:
            df = func(df)
        _logger.debug("python analytics job result:\n%s", df)
        return df

    def get_module_and_func_name(self, config: dict) -> str:
        module_name, func_name = get_profile(self.analytic, config)
        return module_name, func_name


class PythonAnalyticsInterface(AnalyticsInterface):
    def __init__(
        self,
        serialized_cache_catalog: Optional[str] = None,
        session_id: Optional[UUID] = None,
    ):
        _logger.debug("PythonAnalyticsInterface: loading config")
        super().__init__(serialized_cache_catalog, session_id)
        self.config = load_profiles()

    @staticmethod
    def schemes() -> Iterable[str]:
        return ["python"]

    def get_datasources(self) -> List[str]:
        return list(self.config)

    def get_storage_of_datasource(self, datasource: str) -> str:
        return "local"

    def store(
        self,
        instruction_id: UUID,
        data: DataFrame,
    ):
        raise NotImplementedError("PythonAnalyticsInterface.store")  # TEMP

    def explain_graph(
        self,
        graph: IRGraphEvaluable,
        cache: MutableMapping[UUID, Any],
        instructions_to_explain: Optional[Iterable[Instruction]] = None,
    ) -> Mapping[UUID, GraphletExplanation]:
        mapping = {}
        if not instructions_to_explain:
            instructions_to_explain = graph.get_sink_nodes()
        for instruction in instructions_to_explain:
            dep_graph = graph.duplicate_dependent_subgraph_of_node(instruction)
            graph_dict = dep_graph.to_dict()
            job = self._evaluate_instruction_in_graph(graph, cache, instruction)
            module_name, func_name = job.get_module_and_func_name(self.config)
            action = AnalyticOperation("Python", module_name + "::" + func_name)
            mapping[instruction.id] = GraphletExplanation(graph_dict, action)
        return mapping

    def evaluate_graph(
        self,
        graph: IRGraphEvaluable,
        cache: MutableMapping[UUID, Any],
        instructions_to_evaluate: Optional[Iterable[Instruction]] = None,
    ) -> Mapping[UUID, DataFrame]:
        _logger.debug(
            "python: graph: %s", json.dumps(json.loads(graph.to_json()), indent=4)
        )
        mapping = {}
        if not instructions_to_evaluate:
            instructions_to_evaluate = graph.get_sink_nodes()
        for instruction in instructions_to_evaluate:
            _logger.debug("python: inst to evaluate: %s", instruction)
            job = self._evaluate_instruction_in_graph(graph, cache, instruction)
            mapping[job.output_iid] = job.run(self.config)
        return mapping

    def _evaluate_instruction_in_graph(
        self,
        graph: IRGraphEvaluable,
        cache: MutableMapping[UUID, Any],
        instruction: Instruction,
    ) -> PythonAnalyticsJob:
        _logger.debug("python analytics: instruction = %s", instruction)

        if isinstance(instruction, TransformingInstruction):
            if instruction.id in cache:
                iid = instruction.id
                _logger.debug("python analytics: got iid = %s from cache", iid)
                job = PythonAnalyticsJob(iid, cache)
            else:
                trunk, _ = graph.get_trunk_n_branches(instruction)
                job = self._evaluate_instruction_in_graph(graph, cache, trunk)

            if isinstance(instruction, Analytic):
                job.analytic = instruction.name
                job.params = instruction.params
                _logger.debug("python analytics: %s", job.analytic)
            elif isinstance(instruction, Variable):
                if not job.input_iid:
                    job.input_iid = instruction.id
                    _logger.debug("python analytics: input_iid = %s", job.input_iid)
                else:
                    job.output_iid = instruction.id
                    _logger.debug("python analytics: output_iid = %s", job.output_iid)

        return job


class PythonAnalytics(AbstractContextManager):
    def __init__(
        self, analytic_name: str, module_name: str, func_name: str, parameters: dict
    ):
        self.name = analytic_name
        self.func_name = func_name
        self.parameters = parameters
        self.module_path = pathlib.Path(module_name).expanduser().resolve()
        self.module_path_dir_str = str(self.module_path.parent)
        self.use_env = False

    def __enter__(self):
        # accommodate any other Python modules to load in the dir
        self.syspath = sys.path.copy()
        sys.path.append(self.module_path_dir_str)

        # accommodate any other executables or data to load in the dir
        self.cwd_original = os.getcwd()
        os.chdir(self.module_path_dir_str)

        # time to load the analytics function
        self.analytics_function = self._locate_analytics_func(self._load_module())

        # inspect signature of function and skip env vars if it has kwargs
        self.use_env = self._set_env_vars()

        # passing parameters as environment variables
        if self.use_env:
            self.environ_original = os.environ.copy()
            if self.parameters:
                if isinstance(self.parameters, Mapping):
                    parameters = {k: str(v) for k, v in self.parameters.items()}
                    _logger.debug(f"setting parameters as env vars: {parameters}")
                    os.environ.update(parameters)
                else:
                    raise InvalidAnalyticsInterfaceImplementation(
                        "parameters should be passed in as a Mapping"
                    )

        return self._execute

    def __exit__(self, exception_type, exception_value, _traceback):
        sys.path = self.syspath
        os.chdir(self.cwd_original)
        if self.use_env:
            os.environ = self.environ_original

    def _execute(self, dataframe: DataFrame):
        """Execute the analytics

        Args:
            dataframe (DataFrame): input variable to the analytics.

        Returns:

            DataFrame: the analytics output (i.e. "enriched" DataFrame)

        """
        input_dataframes = [dataframe]  # TEMP
        if len(input_dataframes) != self._get_var_count():
            raise InvalidAnalyticsArgumentCount(
                self.name, len(input_dataframes), self._get_var_count()
            )

        try:
            if self.use_env:
                outputs = self.analytics_function(*input_dataframes)
            else:
                outputs = self.analytics_function(*input_dataframes, **self.parameters)
        except Exception as e:
            _logger.error('"%s" failed at execution: %s', self.name, e, exc_info=e)
            raise AnalyticsError(f"{self.name} failed at execution") from e

        if not isinstance(outputs, tuple):
            outputs = (outputs,)

        output_dfs, output_dsps = [], []
        for x in outputs:
            x_class_str = type(x).__module__ + "." + type(x).__name__
            if isinstance(x, DataFrame):
                output_dfs.append(x)
            elif isinstance(x, str):
                _logger.info(
                    f'analytics "{self.name}" yielded a string return. treat it as an HTML element.'
                )
                output_dsps.append(x)  # FIXME:DisplayHtml(x))
            elif x_class_str == "matplotlib.figure.Figure":
                _logger.info(f'analytics "{self.name}" yielded a figure.')
                output_dsps.append(x)  # FIXME:DisplayFigure(x))
            else:
                raise InvalidAnalyticsOutput(self.name, type(x))

        if not outputs:
            raise AnalyticsError(f'analytics "{self.name}" yield nothing')
        if len(output_dsps) > 1:
            raise AnalyticsError(
                f'analytics "{self.name}" yielded more than one Kestrel Display object'
            )
        if output_dfs:
            if len(output_dfs) != len(input_dataframes):
                raise AnalyticsError(
                    f'analytics "{self.name}" yielded less/more Kestrel variable(s) than given'
                )
            return output_dfs[0]  # TEMP

        display = output_dsps[0] if output_dsps else None
        return display

    def _load_module(self):
        spec = spec_from_file_location(
            "kestrel_analytics_python.analytics.{profile_name}", str(self.module_path)
        )

        try:
            module = module_from_spec(spec)
            spec.loader.exec_module(module)
        except ModuleNotFoundError as e:
            raise AnalyticsError(
                f"{self.name} misses dependent library: {e.name}",
                "pip install the corresponding Python package",
            )
        except Exception as e:
            if isinstance(e, AttributeError) and e.args == (
                "'NoneType' object has no attribute 'loader'",
            ):
                raise AnalyticsError(
                    f"{self.name} is not found",
                    "please make sure the Python module and function specified in the profile (configuration) exist",
                )
            else:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                error = "".join(
                    traceback.format_exception(exc_type, exc_value, exc_traceback)
                )
                raise AnalyticsError(f"{self.name} failed at importing:\n{error}")

        return module

    def _locate_analytics_func(self, module):
        if hasattr(module, self.func_name):
            return getattr(module, self.func_name)
        raise InvalidAnalytics(
            self.name,
            "python",
            f'function "{self.func_name}" not exist in module: {self.module_path}',
        )

    def _get_var_count(self):
        """Determine number of vars/DataFrames the analytics func expects"""
        sig = inspect.signature(self.analytics_function)
        # Count of params with type DataFrame
        # If there are no type annotations, then fall back to param count
        df_count = sum(1 for i in sig.parameters.values() if i.annotation == DataFrame)
        return df_count if df_count else len(sig.parameters)

    def _set_env_vars(self):
        """Check if the analytics function DOES NOT accept any non-DataFrame parameters. If so, return True and use env to pass params."""
        sig = inspect.signature(self.analytics_function)
        return (
            sum(1 for i in sig.parameters.values() if i.annotation != DataFrame) == 0
            or len(sig.parameters) == 1
        )



import pytest
from pandas import DataFrame

from kestrel.frontend.parser import parse_kestrel_and_update_irgraph
from kestrel.interface.codegen.dataframe import (
    evaluate_source_instruction, evaluate_transforming_instruction)
from kestrel.ir.graph import IRGraph
from kestrel.ir.instructions import Construct, Limit, ProjectAttrs, Variable
from kestrel.interface.codegen.utils import variable_attributes_to_dataframe


def test_evaluate_Construct():
    data = [ {"name": "cmd.exe", "pid": 123}
           , {"name": "explorer.exe", "pid": 99}
           , {"name": "firefox.exe", "pid": 201}
           , {"name": "chrome.exe", "pid": 205}
           ]
    ins = Construct(data, "process")
    df = evaluate_source_instruction(ins)
    assert df.equals(DataFrame(data))


def test_non_exist_eval():
    with pytest.raises(NotImplementedError):
        evaluate_transforming_instruction(Variable("asdf", "foo", "bar"), DataFrame())


def test_evaluate_Limit():
    data = [ {"name": "cmd.exe", "pid": 123}
           , {"name": "explorer.exe", "pid": 99}
           , {"name": "firefox.exe", "pid": 201}
           , {"name": "chrome.exe", "pid": 205}
           ]
    df = DataFrame(data)
    dfx = evaluate_transforming_instruction(Limit(2), df)
    assert dfx.equals(df.head(2))


def test_evaluate_ProjectAttrs():
    data = [ {"name": "cmd.exe", "pid": 123}
           , {"name": "explorer.exe", "pid": 99}
           , {"name": "firefox.exe", "pid": 201}
           , {"name": "chrome.exe", "pid": 205}
           ]
    df = DataFrame(data)
    dfx = evaluate_transforming_instruction(ProjectAttrs(["name"]), df)
    assert dfx.equals(df[["name"]])


def test_evaluate_Construct_Filter_ProjectAttrs():
    stmt = r"""
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
browsers = proclist WHERE name = 'firefox.exe' OR name = 'chrome.exe'
DISP browsers ATTR name, pid
p2 = proclist WHERE pid > 100
p3 = proclist WHERE name LIKE "c%.exe"
p4 = proclist WHERE name MATCHES r"^c\w{2}\.exe"
"""
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    c = graph.get_nodes_by_type(Construct)[0]
    df0 = evaluate_source_instruction(c)
    assert df0.to_dict("records") == [ {"name": "cmd.exe", "pid": 123}
                                     , {"name": "explorer.exe", "pid": 99}
                                     , {"name": "firefox.exe", "pid": 201}
                                     , {"name": "chrome.exe", "pid": 205}
                                     ]

    browsers = graph.get_variable("browsers")
    ft = next(graph.predecessors(browsers))
    dfx = evaluate_transforming_instruction(ft, df0)
    assert dfx.to_dict("records") == [ {"name": "firefox.exe", "pid": 201}
                                     , {"name": "chrome.exe", "pid": 205}
                                     ]
    proj = next(graph.successors(browsers))
    dfy = evaluate_transforming_instruction(proj, dfx)
    assert dfx.to_dict("records") == [ {"name": "firefox.exe", "pid": 201}
                                     , {"name": "chrome.exe", "pid": 205}
                                     ]

    ft = next(graph.predecessors(graph.get_variable("p2")))
    dfx = evaluate_transforming_instruction(ft, df0)
    assert dfx.to_dict("records") == [ {"name": "cmd.exe", "pid": 123}
                                     , {"name": "firefox.exe", "pid": 201}
                                     , {"name": "chrome.exe", "pid": 205}
                                     ]

    ft = next(graph.predecessors(graph.get_variable("p3")))
    dfx = evaluate_transforming_instruction(ft, df0)
    assert dfx.to_dict("records") == [ {"name": "cmd.exe", "pid": 123}
                                     , {"name": "chrome.exe", "pid": 205}
                                     ]

    ft = next(graph.predecessors(graph.get_variable("p4")))
    dfx = evaluate_transforming_instruction(ft, df0)
    assert dfx.to_dict("records") == [ {"name": "cmd.exe", "pid": 123} ]


def test_information():
    data = [ {"process.name": "cmd.exe", "process.pid": 123, "user.name": "user", "event_type": "process"} ]
    df = DataFrame(data)
    idf = variable_attributes_to_dataframe(df)
    attrs = idf["attributes"].to_list()
    assert attrs == ['event_type', 'process.name, process.pid', 'user.name']



import logging

from kestrel.config.utils import CONFIG_DIR_DEFAULT, load_user_config
from kestrel.exceptions import InvalidAnalytics

PROFILE_PATH_DEFAULT = CONFIG_DIR_DEFAULT / "pythonanalytics.yaml"
PROFILE_PATH_ENV_VAR = "KESTREL_PYTHON_ANALYTICS_CONFIG"

_logger = logging.getLogger(__name__)


def load_profiles():
    config = load_user_config(PROFILE_PATH_ENV_VAR, PROFILE_PATH_DEFAULT)
    if config and "profiles" in config:
        _logger.debug(f"python analytics profiles found in config file")
        profiles = config["profiles"]
    else:
        _logger.info("no python analytics config with profiles found")
        profiles = {}
    _logger.debug(f"profiles loaded: {profiles}")
    return profiles


def get_profile(profile_name, profiles):
    if profile_name not in profiles:
        raise InvalidAnalytics(
            profile_name,
            "python",
            f"no {profile_name} configuration found",
        )
    else:
        profile = profiles[profile_name]
        _logger.debug(f"profile to use: {profile}")
        if "module" not in profile:
            raise InvalidAnalytics(
                profile_name,
                "python",
                f"no {profile_name} module defined",
            )
        else:
            module_name = profile["module"]
        if "func" not in profile:
            raise InvalidAnalytics(
                profile_name,
                "python",
                f"no {profile_name} func defined",
            )
        else:
            func_name = profile["func"]

    return module_name, func_name



import logging
import os

from pandas import DataFrame

_logger = logging.getLogger(__name__)


def do_something(df: DataFrame, **kwargs):
    _logger.debug("python analytics: run pseudo-analytic")
    for k, v in kwargs.items():
        df[k] = v
    return df


def do_something_no_annotations(df):
    _logger.debug("python analytics: run pseudo-analytic with env vars")
    name = os.environ.get("name", "new_column")
    value = int(os.environ.get("value", 0))
    df[name] = value
    return df


def do_something_env(df: DataFrame):
    return do_something_no_annotations(df)



import collections.abc
import os
from datetime import datetime
from importlib import resources
from pathlib import Path
from pkgutil import get_data
from typing import Iterable, Mapping, Optional

from kestrel.__future__ import is_python_older_than_minor_version
from typeguard import typechecked

TIME_FMT = "%Y-%m-%dT%H:%M:%S.%f"


@typechecked
def load_data_file(package_name: str, file_name: str) -> str:
    try:
        # resources.files() is introduced in Python 3.9
        content = resources.files(package_name).joinpath(file_name).read_text()
    except AttributeError:
        # Python 3.8; deprecation warning forward
        if is_python_older_than_minor_version(9):
            content = get_data(package_name, file_name).decode("utf-8")

    return content


@typechecked
def list_folder_files(
    package_name: str,
    folder_name: str,
    prefix: Optional[str] = None,
    extension: Optional[str] = None,
) -> Iterable[str]:
    # preprocesss extension to add dot it not there
    if extension and extension[0] != ".":
        extension = "." + extension
    try:
        file_paths = resources.files(package_name).joinpath(folder_name).iterdir()
    except AttributeError:
        if is_python_older_than_minor_version(9):
            import pkg_resources

            file_names = pkg_resources.resource_listdir(package_name, folder_name)
            file_paths = [
                Path(
                    pkg_resources.resource_filename(
                        package_name, os.path.join(folder_name, filename)
                    )
                )
                for filename in file_names
            ]
    file_list = (
        f
        for f in file_paths
        if (
            f.is_file()
            and (f.name.endswith(extension) if extension else True)
            and (f.name.startswith(prefix) if prefix else True)
        )
    )
    return file_list


@typechecked
def unescape_quoted_string(s: str) -> str:
    if s.startswith("r"):
        return s[2:-1]
    else:
        return s[1:-1].encode("utf-8").decode("unicode_escape")


@typechecked
def update_nested_dict(dict_old: Mapping, dict_new: Optional[Mapping]) -> Mapping:
    if dict_new:
        for k, v in dict_new.items():
            if isinstance(v, collections.abc.Mapping) and k in dict_old:
                dict_old[k] = update_nested_dict(dict_old[k], v)
            else:
                dict_old[k] = v
    return dict_old


@typechecked
def timefmt(t: datetime, prec: int = 3) -> str:
    """Format Python datetime `t` in RFC 3339-format

    Ported from firepit.timestamp
    """
    val = t.strftime(TIME_FMT)
    parts = val.split(".")
    if len(parts) > 1:
        l = len(parts[0])
        digits = parts[1]
        num_digits = len(digits)
        if num_digits:
            l += min(num_digits, prec) + 1
    return val[:l] + "Z"



# parse Kestrel syntax, apply frontend mapping, transform to IR

import logging
from itertools import chain
from typing import Iterable

import lark
import yaml
from kestrel.config.utils import get_all_relations, load_relation_configs
from kestrel.frontend.compile import _KestrelT
from kestrel.ir.graph import IRGraph
from kestrel.ir.instructions import Return
from kestrel.mapping.data_model import reverse_mapping
from kestrel.utils import list_folder_files, load_data_file
from pandas import DataFrame
from typeguard import typechecked

_logger = logging.getLogger(__name__)


MAPPING_MODULE = "kestrel.mapping"

# cache mapping in the module
frontend_mappings = {}

# cache relation table in the module
relation_tables = {}


@typechecked
def get_relation_table(table: str) -> DataFrame:
    global relation_tables
    if table not in relation_tables:
        relation_tables[table] = load_relation_configs(table)
    return relation_tables[table]


@typechecked
def get_frontend_mapping(submodule: str, do_reverse_mapping: bool = False) -> dict:
    global frontend_mappings
    if submodule not in frontend_mappings:
        mapping = {}
        for f in list_folder_files(MAPPING_MODULE, submodule, extension="yaml"):
            with open(f, "r") as fp:
                mapping_ind = yaml.safe_load(fp)
            if do_reverse_mapping:
                mapping_ind = reverse_mapping(mapping_ind)
            mapping.update(mapping_ind)
        frontend_mappings[submodule] = mapping
    return frontend_mappings[submodule]


@typechecked
def get_keywords(including_relations: bool = True):
    grammar = load_data_file("kestrel.frontend", "kestrel.lark")
    parser = lark.Lark(grammar, parser="lalr")
    alphabet_patterns = filter(lambda x: x.pattern.value.isalnum(), parser.terminals)
    all_relations = get_all_relations()
    keywords = [x.pattern.value for x in alphabet_patterns] + all_relations
    keywords_lower = map(lambda x: x.lower(), keywords)
    keywords_upper = map(lambda x: x.upper(), keywords)
    keywords_comprehensive = list(chain(keywords_lower, keywords_upper))
    return keywords_comprehensive


@typechecked
def parse_kestrel_and_update_irgraph(
    stmts: str, irgraph: IRGraph, entity_identifier_map: dict
) -> Iterable[Return]:
    """Parse Kestrel code block and update the input IRGraph

    Parameters:
        stmts: Kestrel code block (statements)
        irgraph: existing IRGraph (used for reference resolution; will be updated)
        entity_identifier_map: identifiers for each entity, required by FIND

    Returns:
        List of Return instructions in the current code block
    """
    lp = lark.Lark(
        load_data_file("kestrel.frontend", "kestrel.lark"),
        parser="lalr",
        transformer=_KestrelT(
            irgraph,
            get_frontend_mapping("fields", True),
            get_frontend_mapping("types"),
            get_relation_table("entity"),
            get_relation_table("event"),
            entity_identifier_map,
        ),
    )
    return lp.parse(stmts)


@typechecked
def parse_without_transform(
    stmts: str,
) -> lark.tree.Tree:
    """Parse Kestrel code block and not transform; for syntax error check"""
    lp = lark.Lark(
        load_data_file("kestrel.frontend", "kestrel.lark"),
        parser="lalr",
    )
    return lp.parse(stmts)



from __future__ import annotations

from abc import abstractmethod
from typing import Iterable, List, MutableMapping
from uuid import UUID

from kestrel.config.internal import CACHE_INTERFACE_IDENTIFIER, CACHE_STORAGE_IDENTIFIER
from kestrel.interface import DatasourceInterface
from pandas import DataFrame


class AbstractCache(DatasourceInterface, MutableMapping):
    """Base class for Kestrel cache"""

    @staticmethod
    def schemes() -> Iterable[str]:
        return [CACHE_INTERFACE_IDENTIFIER]

    def get_datasources(self) -> List[str]:
        return []

    def get_storage_of_datasource(self, datasource: str) -> str:
        return CACHE_STORAGE_IDENTIFIER

    @abstractmethod
    def __del__(self):
        """Delete the cache and release memory/disk space"""
        ...

    @abstractmethod
    def __getitem__(self, instruction_id: UUID) -> DataFrame:
        """Get the dataframe for the cached instruction

        This method will automatically support `uuid in cache`

        Parameters:
            instruction_id: id of the instruction

        Returns:
            dataframe of the given (likely Variable) instruction
        """
        ...

    @abstractmethod
    def __setitem__(self, instruction_id: UUID, data: DataFrame):
        """Store the dataframe of an instruction into cache

        Parameters:

            instruction_id: id of the instruction

            data: data associated with the instruction
        """
        ...

    @abstractmethod
    def __delitem__(self, instruction_id: UUID):
        """Delete cached item

        Parameters:
            instruction_id: id of the instruction
        """
        ...

    @abstractmethod
    def get_virtual_copy(self) -> AbstractCache:
        """Create a virtual cache object from this cache

        This method needs to reimplement __del__, __getitem__, __setitem__,
        __delitem__ to not actually hit the store media, e.g., SQLite.

        The virtual cache is useful for the implementation of the Explain()
        instruction, pretending the dependent graphs are evaluated, so the
        evaluation can continue towards the Return() instruction.

        Because Python invokes special methods from class methods, replacing
        the __getitem__, __setitem__, and __delitem__ in the object does not
        help. It is better to derive a subclass and replace __class__ of the
        object to the subclass to correctly invoke the new set of __xitem___.

        https://docs.python.org/3/reference/datamodel.html#special-lookup

        And Python garbage collector could clean up the virtual cache when
        not in use, so the __del__ method should be reimplemented to make
        sure the store media is not closed.
        """
        ...

    def store(self, instruction_id: UUID, data: DataFrame):
        self[instruction_id] = data

    def __iter__(self) -> UUID:
        """Return UUIDs of instructions cached

        Returns:
            UUIDs in iterator
        """
        return iter(self.cache_catalog)

    def __len__(self) -> int:
        """How many items are cached"""
        return len(self.cache_catalog)



from __future__ import annotations

import copy
import json
import logging
from collections import defaultdict
from itertools import combinations
from typing import Any, Iterable, Mapping, MutableMapping, Optional, Tuple, Union
from uuid import UUID

import networkx
from kestrel.config.internal import CACHE_INTERFACE_IDENTIFIER, CACHE_STORAGE_IDENTIFIER
from kestrel.exceptions import (
    DanglingFilter,
    DanglingReferenceInFilter,
    DataSourceNotFound,
    DuplicatedDataSource,
    DuplicatedReference,
    DuplicatedReferenceInFilter,
    DuplicatedSingletonInstruction,
    DuplicatedVariable,
    InevaluableInstruction,
    InstructionNotFound,
    InvalidSeralizedGraph,
    LargerThanOneIndegreeInstruction,
    MissingReferenceInFilter,
    MultiInterfacesInGraph,
    MultiSourcesInGraph,
    ReferenceNotFound,
    SourceNotFound,
    VariableNotFound,
)
from kestrel.ir.filter import ReferenceValue
from kestrel.ir.instructions import (
    Analytic,
    Construct,
    DataSource,
    Filter,
    Instruction,
    IntermediateInstruction,
    ProjectAttrs,
    Reference,
    Return,
    SolePredecessorTransformingInstruction,
    SourceInstruction,
    TransformingInstruction,
    Variable,
    instruction_from_dict,
)
from typeguard import typechecked

_logger = logging.getLogger(__name__)


@typechecked
def compose(g: IRGraph, h: IRGraph) -> IRGraph:
    g.update(h)
    return g


@typechecked
def union(g: IRGraph, h: IRGraph) -> IRGraph:
    return compose(g, h)


@typechecked
class IRGraph(networkx.DiGraph):
    def __init__(
        self, serialized_graph: Union[None, str, Mapping[str, Iterable[Mapping]]] = None
    ):
        super().__init__()
        if serialized_graph:
            if isinstance(serialized_graph, str):
                graph_in_dict = json.loads(serialized_graph)
            else:
                graph_in_dict = serialized_graph
            self._from_dict(graph_in_dict)

    def add_node(
        self,
        node: Instruction,
        dependent_node: Optional[Instruction] = None,
        deref: bool = True,
    ) -> Instruction:
        """General adding node/instruction operation

        Parameters:
            node: the instruction to add
            dependent_node: the dependent instruction if node is a TransformingInstruction
            deref: whether to dereference Reference instruction

        Returns:
            The node added
        """
        if node not in self:
            if isinstance(node, TransformingInstruction):
                node = self._add_node_with_dependent_node(node, dependent_node)
            else:
                node = self._add_node(node, deref)
        return node

    def add_nodes_from(self, nodes: Iterable[Instruction], deref: bool = True):
        """Add nodes in a list

        Parameters:
            nodes: the list of nodes/instructions to add
            deref: whether to deref when adding node
        """
        for node in nodes:
            self._add_node(node, deref)

    def add_edge(self, u: Instruction, v: Instruction, deref: bool = False):
        """Add edge (add node if not exist)

        Parameters:
            u: the source of the edge
            v: the target of the edge
            deref: whether to deref when adding node
        """
        ux = self._add_node(u, deref)
        vx = self._add_node(v, deref)
        super().add_edge(ux, vx)

    def add_edges_from(
        self, edges: Iterable[Tuple[Instruction, Instruction]], deref: bool = False
    ):
        """Add edges in a list

        Parameters:
            edges: the edges to add
            deref: whether to deref when adding node
        """
        for u, v in edges:
            self.add_edge(u, v, deref)

    def copy(self):
        """Copy the IRGraph with all nodes as reference (not deepcopy)

        Support subclass of IRGraph to be copied.
        """
        g = IRGraph()
        g.update(self)

        # subclass support
        if type(g) != type(self):
            g = type(self)(g)

        return g

    def deepcopy(self):
        """Copy the IRGraph with all nodes copied as new objects

        Support subclass of IRGraph to be deep copied.
        """
        g = IRGraph()
        o2n = {n: copy.deepcopy(n) for n in self.nodes()}
        for u, v in self.edges():
            g.add_edge(o2n[u], o2n[v])
        g.add_nodes_from([o2n[n] for n in self.nodes() if self.degree(n) == 0])

        # subclass support
        if type(g) != type(self):
            g = type(self)(g)

        return g

    def get_node_by_id(self, ux: Union[UUID, str]) -> Instruction:
        """Get node by ID

        Parameters:
            ux: node ID

        Returns:
            The Kestrel instruction (node in IRGraph)
        """
        u = UUID(ux) if isinstance(ux, str) else ux
        try:
            return next(filter(lambda n: n.id == u, self.nodes()))
        except StopIteration:
            raise InstructionNotFound(u)

    def get_nodes_by_type(self, ntype: type) -> Iterable[Instruction]:
        """Get nodes by type

        Parameters:
            ntype: node/instruction type

        Returns:
            The list of nodes/instructions
        """
        return [n for n in self.nodes() if isinstance(n, ntype)]

    def get_nodes_by_type_and_attributes(
        self, ntype: type, attr2val: Mapping[str, Union[str, bool, int]]
    ) -> Iterable[Instruction]:
        """Get nodes by both type and attributes/values

        Parameters:
            ntype: node/instruction type
            attr2val: instruction attribute/value dictionary

        Returns:
            The list of nodes/instructions
        """
        nodes = self.get_nodes_by_type(ntype)
        return [
            n
            for n in nodes
            if all([getattr(n, k, None) == v for (k, v) in attr2val.items()])
        ]

    def get_variable(self, var_name: str) -> Variable:
        """Get a Kestrel variable by its name

        Parameters:
            var_name: variable name

        Returns:
            The Kestrel variable given its name
        """
        xs = self.get_nodes_by_type_and_attributes(Variable, {"name": var_name})
        if xs:
            if len({x.version for x in xs}) < len(xs):
                raise DuplicatedVariable(var_name)
            else:
                xs.sort(key=lambda x: x.version)
                return xs[-1]
        else:
            raise VariableNotFound(var_name)

    def get_variables(self) -> Iterable[Variable]:
        """Get all variables

        This method returns a list of variables, equivalent to *Symbol Table* used in traditional (non-graph-IR) language compilers. Shadowed variables (replaced by new variables with same names) will not be returned.

        Returns:
            The list of all Kestrel variables in this huntflow.
        """
        var_names = {v.name for v in self.get_nodes_by_type(Variable)}
        return [self.get_variable(var_name) for var_name in var_names]

    def get_reference(self, ref_name: str) -> Reference:
        """Get a Kestrel reference by its name

        Parameters:
            ref_name: reference name

        Returns:
            The Reference node
        """
        xs = self.get_nodes_by_type_and_attributes(Reference, {"name": ref_name})
        if xs:
            if len(xs) > 1:
                raise DuplicatedReference(ref_name)
            else:
                return xs.pop()
        else:
            raise ReferenceNotFound(ref_name)

    def get_references(self) -> Iterable[Reference]:
        """Get all references

        Returns:
            The list of reference nodes
        """
        ref_names = {r.name for r in self.get_nodes_by_type(Reference)}
        return [self.get_reference(ref_name) for ref_name in ref_names]

    def add_reference(
        self, rx: Union[str, Reference], deref: bool = True
    ) -> Union[Reference, Variable]:
        """Create or add new reference node to IRGraph

        The reference node will be derefed if the flag is specified.

        Parameters:
            rx: reference name (str) or already created node (Reference)
            deref: whether to deref when adding node

        Returns:
            The reference node created/added
        """
        r = Reference(rx) if isinstance(rx, str) else rx
        return self.add_node(r, deref)

    def get_datasource(self, interface: str, datasource: str) -> DataSource:
        """Get a Kestrel datasource by its URI

        Parameters:
            interface: the datasource interface name
            datasource: the datasource name under the interface

        Returns:
            The datasource
        """
        xs = self.get_nodes_by_type_and_attributes(
            DataSource, {"interface": interface, "datasource": datasource}
        )
        if xs:
            if len(xs) > 1:
                raise DuplicatedDataSource(interface, datasource)
            else:
                return xs.pop()
        else:
            raise DataSourceNotFound(interface, datasource)

    def get_datasources(self) -> Iterable[DataSource]:
        """Get all datasources

        Returns:
            The list of data sources
        """
        xs = self.get_nodes_by_type(DataSource)

        # to check for duplicated datasources

        return xs

    def add_datasource(
        self, sx: Union[str, DataSource], default_interface: Optional[str] = None
    ) -> DataSource:
        """Create new datasource (if needed) and add to IRGraph if not exist

        Parameters:
            sx: the full URI of the datasource (str) or already created node (DataSource)
            default_interface: default interface name

        Returns:
            The DataSource node found or added
        """
        s = DataSource(sx, default_interface) if isinstance(sx, str) else sx
        return self.add_node(s)

    def get_returns(self) -> Iterable[Return]:
        """Get all return nodes

        Returns:
            The list of return nodes
        """
        return sorted(self.get_nodes_by_type(Return), key=lambda x: x.sequence)

    def get_max_return_sequence(self) -> int:
        """Get the largest sequence number of all Returns

        Returns:
            The largest sequence number of all Return instruction
        """
        return max(map(lambda x: x.sequence, self.get_returns()), default=-1)

    def add_return(self, dependent_node: Instruction) -> Return:
        """Create new Return instruction and add to IRGraph

        Parameters:
            dependent_node: the instruction to hold return

        Returns:
            The return node created/added
        """
        return self.add_node(Return(), dependent_node)

    def get_sink_nodes(self) -> Iterable[Instruction]:
        """Get all sink nodes (node with no successors)

        Returns:
            The list of sink nodes
        """
        return [n for n in self.nodes() if self.out_degree(n) == 0]

    def get_trunk_n_branches(
        self, node: TransformingInstruction
    ) -> (Instruction, Mapping[ReferenceValue, Instruction]):
        """Get the trunk and branches paths for instruction

        For trunk path, return the tail node; for each branch, return the tail
        node of the branchin mapping from reference to node.

        Parameters:
            node: the instruction node

        Returns:
            (tail node for trunk, ref to branch tail node mapping)
        """
        ps = list(self.predecessors(node))
        pps = [(p, pp) for p in self.predecessors(node) for pp in self.predecessors(p)]

        if not ps:
            raise SourceNotFound(node)

        # may need to add a patch in find_dependent_subgraphs_of_node()
        # for each new case added in the if/elif, e.g., FIlter
        if isinstance(node, SolePredecessorTransformingInstruction):
            if len(ps) > 1:
                raise LargerThanOneIndegreeInstruction()
            else:
                return ps[0], {}
        elif isinstance(node, Filter):
            r2n = {}
            for rv in node.get_references():
                ppfs = [
                    (p, pp)
                    for p, pp in pps
                    if isinstance(p, ProjectAttrs)
                    and isinstance(pp, (Variable, Reference))
                    and p.attrs == rv.attributes
                    and pp.name == rv.reference
                ]
                if not ppfs:
                    raise MissingReferenceInFilter(rv, node, pps)
                elif len(ppfs) > 1:
                    raise DuplicatedReferenceInFilter(rv, node, pps)
                else:
                    p = ppfs[0][0]
                    r2n[rv] = p
                    ps.remove(p)
            if len(ps) == 0:
                raise DanglingFilter()
            elif len(ps) > 1:
                raise DanglingReferenceInFilter(ps)
            return ps[0], r2n
        elif isinstance(node, Analytic):
            ps = [x for x in ps if isinstance(x, (Variable, Reference))]
            return ps[0], {}
        else:
            # TODO: x = y + z, which is trunk?
            raise NotImplementedError(f"unknown instruction type: {node}")

    def update(self, ng: IRGraph):
        """Extend the current IRGraph with a new IRGraph

        Parameters:
            ng: the new IRGraph to merge/combine/union
        """
        # After we add new variable nodes, we can no longer rely on
        # self.get_variable() to get variables for de-referencing.
        # Save the original variables first.
        original_variables = {v.name: v for v in self.get_variables()}

        # prepare new variables from ng before merge
        # should not use ng.get_variable(),
        # which does not cover all overridden variables
        for nv in ng.get_nodes_by_type(Variable):
            if nv.name in original_variables:
                nv.version += original_variables[nv.name].version + 1

        # prepare return sequence from ng before merge
        return_max_sequence = self.get_max_return_sequence()
        for nr in ng.get_returns():
            nr.sequence += return_max_sequence + 1

        # add refs first to deref correctly
        # if any reference exist, it should be derefed before adding any variable
        o2n_refs = {n: self._add_node(n) for n in ng.get_references()}
        # add all nodes with dedup singleton node, e.g., SourceInstruction
        o2n_nonrefs = {n: self._add_node(n) for n in ng.nodes() if n not in o2n_refs}

        # overall old to new node mapping
        o2n = {}
        o2n.update(o2n_refs)
        o2n.update(o2n_nonrefs)

        # add all edges
        self.add_edges_from([(o2n[u], o2n[v]) for (u, v) in ng.edges()])

    def duplicate_dependent_subgraph_of_node(self, node: Instruction) -> IRGraph:
        """Find and copy the dependent subgraph of a node (including the node)

        Parameters:
            node: instruction node to start

        Returns:
            A copy of the dependent subgraph (including the input node)
        """
        nodes = networkx.ancestors(self, node)
        nodes.add(node)
        return self.subgraph(nodes).copy()

    def find_cached_dependent_subgraph_of_node(
        self, node: Instruction, cache: MutableMapping[UUID, Any]
    ) -> IRGraph:
        """Return the cached dependent graph of the a node

        Discard nodes and subgraphs before any cached nodes, e.g., Variables.

        Parameters:
            node: instruction node to start
            cache: any type of node cache, e.g., content, SQL statement

        Returns:
            The pruned IRGraph without nodes before cached Variable nodes
        """
        g = self.duplicate_dependent_subgraph_of_node(node)
        in_edges = [g.in_edges(n) for n in g.nodes() if n.id in cache]
        g.remove_edges_from(set().union(*in_edges))

        # important last step to discard any unconnected nodes/subgraphs prior to the dropped edges
        return g.duplicate_dependent_subgraph_of_node(node)

    def find_dependent_subgraphs_of_node(
        self,
        node: Instruction,
        cache: MutableMapping[UUID, Any],
    ) -> Iterable[IRGraphEvaluable]:
        """Find dependency subgraphs that do not have further dependency

        To evaluate a node, one needs to evaluate all nodes in its dependent
        graph. However, not all nodes can be evaluated at once (e.g., impacted
        by multiple interfaces). Some require more basic dependent subgraphs to
        be evaluated first. This method segments the dependent graph of a node
        and return the subgraphs that are IRGraphEvaluable. One can evaluate
        the returns, cache them, and call this method again. After iterations
        of return and evaluation of returned dependent subgraphs, the node can
        finally be evaluated in the last return, which will just be a
        IRGraphEvaluable at that time.

        TODO: [optimization] consider trunk or branch to reduce the
        subgraphs---trunk does not need segmentation

        Parameters:
            node: the instruction/node to generate dependent subgraphs for
            cache: any type of node cache, e.g., content, SQL statement

        Returns:
            A list of subgraphs that do not have further dependency
        """
        _CII = (CACHE_INTERFACE_IDENTIFIER, CACHE_STORAGE_IDENTIFIER)

        # the base graph to segment
        g = self.find_cached_dependent_subgraph_of_node(node, cache)

        # Mapping: {interface name: [impacted nodes]}
        a2ns = defaultdict(set)
        for n in g.get_nodes_by_type(SourceInstruction):
            a2ns[(n.interface, n.store)].add(n)
            a2ns[(n.interface, n.store)].update(networkx.descendants(g, n))

        # all predecessor nodes to any interface impacted nodes
        pns = set().union(*[set(g.predecessors(n)) for ns in a2ns.values() for n in ns])

        # add non-source nodes to cache as default execution environment
        # e.g., a path starting from a cached Variable
        # cached nodes directly preceeding an interface impacted node do not need evaluation
        # they will be given to the interface in cache directly
        cached_nodes = set([n for n in g.nodes() if n.id in cache])
        for n in cached_nodes - pns:
            a2ns[_CII].add(n)
            a2ns[_CII].update(networkx.descendants(g, n))

        # find all nodes that are affected by two or more interfaces
        shared_impacted_nodes = set().union(
            *[a2ns[ix] & a2ns[iy] for ix, iy in combinations(a2ns.keys(), 2)]
        )

        # unshared nodes for each interface
        a2uns = {k: v - shared_impacted_nodes for k, v in a2ns.items()}

        # handle direct predecessor node cached
        # such nodes are required in building dep graphs around interfaces
        # such nodes could be shared by multiple interfaces; can only handle now
        for k in set(a2uns) - set([_CII]):
            ps = set().union(*[set(g.predecessors(n)) for n in a2uns[k]])
            a2uns[k].update(ps & cached_nodes)

        # Add Variable/Reference node if succeeded by ProjectAttrs and Filter,
        # which will be needed (for metadata) by get_trunk_n_branches() in the
        # evaluation phase, through the succeeding node (ProjectAttrs) is
        # already in cache.
        for k in a2uns:
            auxs = []
            for n in a2uns[k]:
                if isinstance(n, ProjectAttrs):
                    # need to search in `self`, not `g`, since the boundry of
                    # `g` is cut by the cache and ProjectAttrs will be cached
                    p = next(self.predecessors(n))
                    s = next(g.successors(n))
                    if (
                        isinstance(s, Filter)
                        and isinstance(p, (Variable, Reference))
                        and s in a2uns[k]
                    ):
                        auxs.append(p)
            a2uns[k].update(auxs)

        # remove dep graphs with only one node
        # e.g., `ds://a` in "y = GET file FROM ds://a WHERE x = v.x"
        # when v.x not in cache (evaluation needed for v.x first)
        dep_nodes = [ns for ns in a2uns.values() if len(ns) > 1]
        # need to search in `self` due to the patch for get_trunk_n_branches()
        dep_graphs = [
            IRGraphEvaluable(self.subgraph(ns)).deepcopy() for ns in dep_nodes
        ]

        return dep_graphs

    def find_simple_query_subgraphs(
        self, cache: MutableMapping[UUID, Any]
    ) -> Iterable[IRGraphSimpleQuery]:
        """Find dependency subgraphs those are IRGraphSimpleQuery

        Some interfaces, e.g., stix-shifter, build stateless query and do not
        support JOIN or sub query/SELECT, so they can only evaluate a simple
        SQL query around each source node. Use this method to prepare such tiny
        graph segments for evaluation by the interface. The remaining of the
        graph can be evaluated in cache.

        Parameters:
            cache: any type of node cache, e.g., content, SQL statement

        Returns:
            An iterator of simple-query subgraphs
        """
        for n in self.get_nodes_by_type(SourceInstruction):
            for g in self._find_paths_from_node_to_a_variable(n, cache):
                yield IRGraphSimpleQuery(g)

    def _find_paths_from_node_to_a_variable(
        self, node: Instruction, cache: MutableMapping[UUID, Any]
    ) -> Iterable[IRGraph]:
        """Find paths (linear IRGraph with directly attached cached nodes) from
        the starting node to its closest variables

        If the linear IRGraph has a dependent branch/path longer than a cached
        node, this linear IRGraph cannot be used to build a IRGraphSimpleQuery;
        it needs to generate a subquery for the branch.

        Parameters:
            node: the node to start path search
            cache: any type of node cache, e.g., content, SQL statement

        Returns:
            An iterator of paths
        """
        # check whether the node has other uncached incoming nodes
        # if no, this path can be a IRGraphSimpleQuery
        if len([n for n in self.predecessors(node) if n.id not in cache]) <= 1:
            # pcns: predecessor cached nodes
            pcns = [n for n in self.predecessors(node) if n.id in cache]
            for succ in self.successors(node):
                if isinstance(succ, Variable):
                    yield self.subgraph([succ, node] + pcns)
                else:
                    for succ_graph in self._find_paths_from_node_to_a_variable(
                        succ, cache
                    ):
                        yield self.subgraph(list(succ_graph.nodes()) + [node] + pcns)

    def to_dict(self) -> Mapping[str, Iterable[Mapping]]:
        """Serialize to a Python dictionary (D3 graph format)

        Returns:
            The graph in a Python dictionary to be dumped as JSON string
        """
        nodes = [n.to_dict() for n in self.nodes()]
        links = [{"source": str(u.id), "target": str(v.id)} for (u, v) in self.edges()]
        return {"nodes": nodes, "links": links}

    def to_json(self) -> str:
        """Serialize to a Python JSON string (D3 graph format)

        Returns:
            The graph in a Python JSON string
        """
        return json.dumps(self.to_dict())

    def find_datasource_of_node(
        self, start: TransformingInstruction
    ) -> Union[DataSource, Construct]:
        """Search for the DataSource of the variable

        Note that AnalyticsInterface is not the trunk of the Analytic node, so
        it should not be a return type here.

        Parameters:
            node: the node to start search

        Returns:
            DataSource of the node
        """
        node = start
        while isinstance(node, TransformingInstruction):
            node, _ = self.get_trunk_n_branches(node)
        if not isinstance(node, (DataSource, Construct)):
            raise SourceNotFound(node)
        else:
            return node

    def _add_node(self, node: Instruction, deref: bool = True) -> Instruction:
        """Add just the node

        Dependency (if exists) not handled. Variable version and Return
        sequence intentionally not handled here (handled in
        _add_node_with_dependent_node()) for plain adding node opeartion used
        by update().

        Parameters:
            node: the node/instruction to add
            deref: whether to deref a Reference node

        Returns:
            The node added or found or derefed
        """
        # test `node in self` is important
        # there could be a Reference node already in graph, not to deref
        if node not in self:
            if isinstance(node, IntermediateInstruction):
                if isinstance(node, Reference):
                    if deref:
                        try:
                            v = self.get_variable(node.name)
                        except VariableNotFound:
                            # deref failed, add Reference node directly
                            node = self._add_singleton_instruction(node)
                        else:
                            # deref succeed, no need to add node
                            node = v
                    else:
                        node = self._add_singleton_instruction(node)
                else:
                    raise NotImplementedError(
                        f"unknown IntermediateInstruction: {node}"
                    )
            elif isinstance(node, SourceInstruction):
                node = self._add_singleton_instruction(node)
            else:
                super().add_node(node)
        return node

    def _add_singleton_instruction(self, node: Instruction) -> Instruction:
        """Guard adding a singleton node

        1. Singleton nodes are nodes that only has one copy in graph

        2. A node that has no predecessors is a singleton node

        Parameters:
            node: the node/instruction to add

        Returns:
            The node added or found
        """
        xs = [
            x
            for x in self.nodes()
            if x.has_same_content_as(node) and self.in_degree(x) == 0
        ]
        if xs:
            if len(xs) > 1:
                raise DuplicatedSingletonInstruction(node)
            else:
                node = xs.pop()
        else:
            super().add_node(node)
        return node

    def _add_node_with_dependent_node(
        self, node: Instruction, dependent_node: Instruction
    ) -> Instruction:
        """Add node to graph with a dependent node

        Variable version and Return sequence are handled here.
        Implicit dependencies (referred variable) of Filter are handled here.

        Parameters:
            node: the node/instruction to add
            dependent_node: the dependent node that should exist in the graph

        Return:
            The node added
        """
        if dependent_node not in self:
            raise InstructionNotFound(dependent_node)
        if node not in self:
            # version control of Variable and Return
            if isinstance(node, Variable):
                try:
                    ve = self.get_variable(node.name)
                except VariableNotFound:
                    node.version = 0
                else:
                    node.version = ve.version + 1
            if isinstance(node, Return):
                node.sequence = self.get_max_return_sequence() + 1

            # add_edge will add node first
            self.add_edge(dependent_node, node)

            # Reference handling for filter
            if isinstance(node, Filter):
                for refvalue in node.get_references():
                    r = self.add_node(Reference(refvalue.reference))
                    p = self.add_node(ProjectAttrs(refvalue.attributes), r)
                    self.add_edge(p, node)
        return node

    def _from_dict(self, graph_in_dict: Mapping[str, Iterable[Mapping]]):
        """Deserialize from a Python dictionary (D3 graph format)

        This method is an implicit constructor from a serialized graph.

        Parameters:
            graph_in_dict: the serialized graph in Python dictionary
        """
        nodes = graph_in_dict["nodes"]
        edges = graph_in_dict["links"]
        for n in nodes:
            self._add_node(instruction_from_dict(n), False)
        for e in edges:
            try:
                u = self.get_node_by_id(e["source"])
                v = self.get_node_by_id(e["target"])
            except InstructionNotFound as err:
                raise InvalidSeralizedGraph()
            else:
                self.add_edge(u, v)


@typechecked
class IRGraphEvaluable(IRGraph):
    """Evaluable IRGraph

    An evaluable IRGraph is an IRGraph that

        1. Only has one interface

        2. Only has one store

        3. No IntermediateInstruction node
    """

    def __init__(self, graph: Optional[IRGraph] = None):
        super().__init__()

        # need to initialize it before `self.update(graph)` below
        self.interface = None
        self.store = None

        # update() will call _add_node() internally to set self.interface
        if graph:
            self.update(graph)

        # all source nodes are already cached (no SourceInstruction)
        if not self.interface:
            self.interface = CACHE_INTERFACE_IDENTIFIER
            self.store = CACHE_STORAGE_IDENTIFIER

    def _add_node(self, node: Instruction, deref: bool = True) -> Instruction:
        if isinstance(node, IntermediateInstruction):
            raise InevaluableInstruction(node)
        elif isinstance(node, SourceInstruction):
            if self.interface:
                if node.interface != self.interface or node.store != self.store:
                    raise MultiInterfacesInGraph(
                        [
                            "INTERFACE: ",
                            self.interface,
                            node.interface,
                            "STORE: ",
                            self.store,
                            node.store,
                        ]
                    )
            else:
                self.interface = node.interface
                self.store = node.store
        return super()._add_node(node, deref)

    def to_dict(self) -> Mapping[str, Iterable[Mapping]]:
        d = super().to_dict()
        d["interface"] = self.interface
        d["store"] = self.store
        return d


@typechecked
class IRGraphSimpleQuery(IRGraphEvaluable):
    """Simple Query IRGraph

    A simple query IRGraph is an evaluable IRGraph that

        1. It contains one source node

        2. It can be compiled into a simple (not nested/joined) SQL query
    """

    def __init__(self, graph: Optional[IRGraph] = None):
        if graph and len(graph.get_nodes_by_type(SourceInstruction)) > 1:
            raise MultiSourcesInGraph()
        super().__init__(graph)



from copy import copy
from typing import Any, Iterable, Mapping, MutableMapping, Optional
from uuid import UUID

from kestrel.cache.base import AbstractCache
from kestrel.display import GraphletExplanation, NativeQuery
from kestrel.interface.codegen.dataframe import (
    evaluate_source_instruction,
    evaluate_transforming_instruction,
)
from kestrel.ir.graph import IRGraphEvaluable
from kestrel.ir.instructions import (
    Explain,
    Filter,
    Instruction,
    Return,
    SourceInstruction,
    TransformingInstruction,
    Variable,
)
from pandas import DataFrame
from typeguard import typechecked


@typechecked
class InMemoryCache(AbstractCache):
    def __init__(
        self,
        initial_cache: Mapping[UUID, DataFrame] = {},
        session_id: Optional[UUID] = None,
    ):
        super().__init__(session_id)
        self.cache: MutableMapping[UUID, DataFrame] = {}

        # update() will call __setitem__() internally
        self.update(initial_cache)

    def __del__(self):
        del self.cache

    def __getitem__(self, instruction_id: UUID) -> DataFrame:
        return self.cache[self.cache_catalog[instruction_id]]

    def __delitem__(self, instruction_id: UUID):
        del self.cache[self.cache_catalog[instruction_id]]
        del self.cache_catalog[instruction_id]

    def __setitem__(
        self,
        instruction_id: UUID,
        data: DataFrame,
    ):
        self.cache_catalog[instruction_id] = instruction_id.hex
        self.cache[self.cache_catalog[instruction_id]] = data

    def get_virtual_copy(self) -> AbstractCache:
        v = copy(self)
        v.cache_catalog = copy(self.cache_catalog)
        v.cache = copy(self.cache)
        return v

    def evaluate_graph(
        self,
        graph: IRGraphEvaluable,
        cache: MutableMapping[UUID, Any],
        instructions_to_evaluate: Optional[Iterable[Instruction]] = None,
    ) -> Mapping[UUID, DataFrame]:
        mapping = {}
        if not instructions_to_evaluate:
            instructions_to_evaluate = graph.get_sink_nodes()
        for instruction in instructions_to_evaluate:
            df = self._evaluate_instruction_in_graph(graph, instruction)
            self[instruction.id] = df
            mapping[instruction.id] = df
        return mapping

    def explain_graph(
        self,
        graph: IRGraphEvaluable,
        instructions_to_explain: Optional[Iterable[Instruction]] = None,
    ) -> Mapping[UUID, GraphletExplanation]:
        mapping = {}
        if not instructions_to_explain:
            instructions_to_explain = graph.get_sink_nodes()
        for instruction in instructions_to_explain:
            dep_graph = graph.duplicate_dependent_subgraph_of_node(instruction)
            graph_dict = dep_graph.to_dict()
            query = NativeQuery("DataFrame", "")
            mapping[instruction.id] = GraphletExplanation(graph_dict, query)
        return mapping

    def _evaluate_instruction_in_graph(
        self, graph: IRGraphEvaluable, instruction: Instruction
    ) -> DataFrame:
        if instruction.id in self:
            df = self[instruction.id]
        elif isinstance(instruction, SourceInstruction):
            df = evaluate_source_instruction(instruction)
        elif isinstance(instruction, TransformingInstruction):
            trunk, r2n = graph.get_trunk_n_branches(instruction)
            df = self._evaluate_instruction_in_graph(graph, trunk)
            if isinstance(instruction, (Return, Explain)):
                pass
            elif isinstance(instruction, Variable):
                self[instruction.id] = df
            else:
                if isinstance(instruction, Filter):
                    # replace each ReferenceValue with a list of values
                    instruction.resolve_references(
                        lambda x: self._evaluate_instruction_in_graph(graph, r2n[x])
                    )
                df = evaluate_transforming_instruction(instruction, df)
        else:
            raise NotImplementedError(f"Unknown instruction type: {instruction}")
        return df



from pathlib import PurePosixPath, PureWindowsPath


class Path:
    """Wrapper class for handling file paths from either Windows or POSIX systems"""

    def __init__(self, raw_path: str):
        """Create a path object that respects the original path separator"""
        if "\\" in raw_path:
            self._path = PureWindowsPath(raw_path)
        elif "/" in raw_path:
            self._path = PurePosixPath(raw_path)
        else:
            # need some heuristics to guess the path type
            self._path = PureWindowsPath(raw_path)  # TODO: more advanced detection?

    def basename(self) -> str:
        """Returns the path with any leading directories removed"""
        return str(self._path.name)

    def dirname(self) -> str:
        """Returns the path with the last component removed, or "." if there's only 1 component"""
        return str(self._path.parent)



from __future__ import annotations

import inspect
import json
import sys
import uuid
from dataclasses import InitVar, dataclass, field, fields
from enum import Enum
from io import StringIO
from typing import Any, Callable, Iterable, Mapping, Optional, Tuple, Type, Union

from kestrel.__future__ import is_python_older_than_minor_version
from kestrel.config.internal import CACHE_INTERFACE_IDENTIFIER, CACHE_STORAGE_IDENTIFIER
from kestrel.exceptions import (
    InvalidDataSource,
    InvalidInstruction,
    InvalidSeralizedInstruction,
)
from kestrel.ir.filter import (
    AbsoluteTrue,
    FExpression,
    ReferenceValue,
    TimeRange,
    get_references_from_exp,
    resolve_reference_with_function,
)
from mashumaro.mixins.json import DataClassJSONMixin
from mashumaro.types import SerializableType
from pandas import DataFrame, read_json
from typeguard import typechecked

# https://stackoverflow.com/questions/70400639/how-do-i-get-python-dataclass-initvar-fields-to-work-with-typing-get-type-hints
if is_python_older_than_minor_version(11):
    InitVar.__call__ = lambda *args: None


class SerializableDataFrame(DataFrame, SerializableType):
    def _serialize(self):
        return self.to_json()

    @classmethod
    def _deserialize(cls, json_str):
        return read_json(StringIO(json_str))

    def __copy__(self):
        return SerializableDataFrame(super().__copy__())

    def __deepcopy__(self, *args):
        return SerializableDataFrame(super().__deepcopy__(*args))


@dataclass
class Instruction(DataClassJSONMixin):
    id: uuid.UUID = field(init=False)
    instruction: str = field(init=False)

    def __post_init__(self):
        # stable id during Instruction lifetime
        self.id = uuid.uuid4()
        self.instruction = self.__class__.__name__

    def __eq__(self, other: Instruction):
        return self.id == other.id

    def __hash__(self):
        # stable hash during Instruction lifetime
        return self.id.int

    def has_same_content_as(self, instruction: Instruction) -> bool:
        if self.instruction == instruction.instruction:
            flag = True
            for f in fields(self):
                if f.name not in ("id", "instruction", "store"):
                    self_data = getattr(self, f.name)
                    other_data = getattr(instruction, f.name)
                    if isinstance(self_data, DataFrame):
                        flag &= self_data.equals(other_data)
                    else:
                        flag &= self_data == other_data
        else:
            flag = False
        return flag


class TransformingInstruction(Instruction):
    """The instruction that builds/dependent on one or more instructions"""

    pass


class SolePredecessorTransformingInstruction(TransformingInstruction):
    """The translating instruction whose indegree==1"""

    pass


class SourceInstruction(Instruction):
    """The instruction that does not dependent on any instruction"""

    interface: str


class IntermediateInstruction(Instruction):
    """The instruction that aids AST to Kestrel IR compilation"""

    pass


@dataclass(eq=False)
class Return(SolePredecessorTransformingInstruction):
    """The sink instruction that forces execution

    Return is implemented as a TransformingInstruction so it triggers
    IRGraph._add_node_with_dependent_node() in IRGraph.add_node()
    """

    # the order/sequence of return instruction in huntflow (source code)
    sequence: int = 0


@dataclass(eq=False)
class Filter(TransformingInstruction):
    exp: FExpression = AbsoluteTrue()
    timerange: TimeRange = field(default_factory=TimeRange)

    # TODO: from_json() for self.exp

    def get_references(self) -> Iterable[ReferenceValue]:
        return get_references_from_exp(self.exp)

    def resolve_references(self, f: Callable[[ReferenceValue], Any]):
        resolve_reference_with_function(self.exp, f)


@dataclass(eq=False)
class ProjectEntity(SolePredecessorTransformingInstruction):
    ocsf_field: str
    native_field: str


@dataclass(eq=False)
class ProjectAttrs(SolePredecessorTransformingInstruction):
    # mashumaro does not support typing.Iterable, only List/Tuple
    attrs: Tuple[str]


@dataclass(eq=False)
class DataSource(SourceInstruction):
    uri: InitVar[Optional[str]] = None
    default_interface: InitVar[Optional[str]] = None
    interface: str = ""
    datasource: str = ""

    # additional info; mapped from self.datasource
    # not used to decide whether two DataSource instructions are the same
    store: str = ""

    def __post_init__(self, uri: Optional[str], default_interface: Optional[str]):
        super().__post_init__()
        if uri:
            # normal constructor, not from deserliazation
            xs = uri.split("://")
            if len(xs) == 2:
                self.interface = xs[0]
                self.datasource = xs[1]
            elif len(xs) == 1 and default_interface:
                self.interface = default_interface
                self.datasource = xs[0]
            else:
                raise InvalidDataSource(uri)
        else:
            # from deserliazation; mashumaro will take care
            pass


@dataclass(eq=False)
class AnalyticsInterface(SourceInstruction):
    interface: str
    store = None


@dataclass(eq=False)
class Analytic(TransformingInstruction):
    name: str
    params: Mapping[str, Union[str, int, float, bool]]


@dataclass(eq=False)
class Variable(SolePredecessorTransformingInstruction):
    name: str
    entity_type: str
    native_type: str
    # required to dereference a variable that has been created multiple times
    # the variable with the largest version will be used by dereference
    version: int = 0


@dataclass(eq=False)
class Reference(IntermediateInstruction):
    """Referred Kestrel variable (used in AST) before de-referencing to a Kestrel variable"""

    name: str


@dataclass(eq=False)
class Explain(SolePredecessorTransformingInstruction):
    pass


@dataclass(eq=False)
class Information(SolePredecessorTransformingInstruction):
    pass


@dataclass(eq=False)
class Limit(SolePredecessorTransformingInstruction):
    num: int


@dataclass(eq=False)
class Offset(SolePredecessorTransformingInstruction):
    num: int


@dataclass(eq=False)
class Construct(SourceInstruction):
    data: SerializableDataFrame
    entity_type: Optional[str] = None
    interface: str = CACHE_INTERFACE_IDENTIFIER
    store: str = CACHE_STORAGE_IDENTIFIER

    def __post_init__(self):
        if type(self.data) != SerializableDataFrame:
            self.data = SerializableDataFrame(self.data)
        super().__post_init__()


class SortDirection(str, Enum):
    ASC = "ASC"
    DESC = "DESC"


@dataclass(eq=False)
class Sort(SolePredecessorTransformingInstruction):
    attribute: str
    direction: SortDirection = SortDirection.DESC


@typechecked
def get_instruction_class(name: str) -> Type[Instruction]:
    classes = inspect.getmembers(sys.modules[__name__], inspect.isclass)
    instructions = [cls for _, cls in classes if issubclass(cls, Instruction)]
    try:
        return next(filter(lambda cls: cls.__name__ == name, instructions))
    except StopIteration:
        raise InvalidInstruction(name)


@typechecked
def instruction_from_dict(d: Mapping[str, Union[str, bool, int]]) -> Instruction:
    instruction_class = get_instruction_class(d["instruction"])
    try:
        instruction = instruction_class.from_dict(d)
        instruction.id = uuid.UUID(d["id"])
    except:
        raise InvalidSeralizedInstruction(d)
    else:
        return instruction


@typechecked
def instruction_from_json(json_str: str) -> Instruction:
    instruction_in_dict = json.loads(json_str)
    return instruction_from_dict(instruction_in_dict)



from itertools import groupby
from typing import List

from pandas import DataFrame


def variable_attributes_to_dataframe(attrs: List[str]) -> DataFrame:
    categories = []
    for k, g in groupby(sorted(attrs), lambda s: s.split(".")[0] if "." in s else ""):
        categories.append(", ".join(g))
    return DataFrame(data={"attributes": categories})



class KestrelError(Exception):
    pass


class InstructionNotFound(KestrelError):
    pass


class InvalidInstruction(KestrelError):
    pass


class InvalidComparison(KestrelError):
    pass


class MismatchedFieldValueInMultiColumnComparison(KestrelError):
    pass


class InvalidOperatorInMultiColumnComparison(KestrelError):
    pass


class InvalidSeralizedGraph(KestrelError):
    pass


class InvalidSeralizedInstruction(KestrelError):
    pass


class InvalidDataSource(KestrelError):
    pass


class InvalidYamlInConfig(KestrelError):
    pass


class InvalidKestrelConfig(KestrelError):
    pass


class MissingEntityIdentifierInConfig(KestrelError):
    pass


class InvalidKestrelRelationTable(KestrelError):
    pass


class UnsupportedObjectRelation(KestrelError):
    pass


class DuplicatedRelationMapping(KestrelError):
    pass


class VariableNotFound(KestrelError):
    pass


class SourceNotFound(KestrelError):
    pass


class ReferenceNotFound(KestrelError):
    pass


class DataSourceNotFound(KestrelError):
    pass


class DuplicatedVariable(KestrelError):
    pass


class DuplicatedReference(KestrelError):
    pass


class DuplicatedDataSource(KestrelError):
    pass


class DuplicatedSingletonInstruction(KestrelError):
    pass


class MultiInterfacesInGraph(KestrelError):
    pass


class MultiSourcesInGraph(KestrelError):
    pass


class LargerThanOneIndegreeInstruction(KestrelError):
    pass


class DanglingReferenceInFilter(KestrelError):
    pass


class DanglingFilter(KestrelError):
    pass


class DuplicatedReferenceInFilter(KestrelError):
    pass


class MissingReferenceInFilter(KestrelError):
    pass


class InvalidSerializedDatasourceInterfaceCacheCatalog(KestrelError):
    pass


class InevaluableInstruction(KestrelError):
    pass


class MappingParseError(KestrelError):
    pass


class InterfaceNotFound(KestrelError):
    pass


class IRGraphMissingNode(KestrelError):
    pass


class InterfaceNotConfigured(KestrelError):
    pass


class InvalidInterfaceImplementation(KestrelError):
    pass


class ConflictingInterfaceScheme(KestrelError):
    pass


class DataSourceError(KestrelError):
    pass


class UnsupportedOperatorError(KestrelError):
    """The data source doesn't support this operator"""

    pass


class IncompleteDataMapping(KestrelError):
    pass


class InvalidAnalytics(KestrelError):
    pass


class InvalidAnalyticsArgumentCount(KestrelError):
    pass


class InvalidAnalyticsInterfaceImplementation(KestrelError):
    pass


class InvalidAnalyticsOutput(KestrelError):
    pass


class AnalyticsError(KestrelError):
    pass


class SourceSchemaNotFound(KestrelError):
    pass


class InvalidProjectEntityFromEntity(KestrelError):
    pass


class EntityNotFound(KestrelError):
    pass


class InvalidMappingWithMultipleIdentifierFields(KestrelError):
    pass


class InvalidTransformerInMapping(KestrelError):
    pass


class InvalidAttributes(KestrelError):
    pass



import functools
import inspect
import operator
import re
import sys
from typing import Callable

from kestrel.exceptions import (
    InvalidAttributes,
    InvalidOperatorInMultiColumnComparison,
    MismatchedFieldValueInMultiColumnComparison,
)
from kestrel.interface.codegen.utils import variable_attributes_to_dataframe
from kestrel.ir.filter import (
    AbsoluteTrue,
    BoolExp,
    ExpOp,
    FBasicComparison,
    FExpression,
    ListOp,
    MultiComp,
    NumCompOp,
    RefComparison,
    StrCompOp,
)
from kestrel.ir.instructions import (
    Construct,
    Filter,
    Information,
    Limit,
    ProjectAttrs,
    ProjectEntity,
    SourceInstruction,
    TransformingInstruction,
)
from pandas import DataFrame, Series
from typeguard import typechecked


@typechecked
def evaluate_source_instruction(instruction: SourceInstruction) -> DataFrame:
    eval_func = _select_eval_func(instruction.instruction)
    return eval_func(instruction)


@typechecked
def evaluate_transforming_instruction(
    instruction: TransformingInstruction, dataframe: DataFrame
) -> DataFrame:
    eval_func = _select_eval_func(instruction.instruction)
    return eval_func(instruction, dataframe)


@typechecked
def _select_eval_func(instruction_name: str) -> Callable:
    eval_funcs = inspect.getmembers(sys.modules[__name__], inspect.isfunction)
    try:
        _funcs = filter(lambda x: x[0] == "_eval_" + instruction_name, eval_funcs)
        return next(_funcs)[1]
    except StopIteration:
        raise NotImplementedError(
            f"evaluation function for {instruction_name} in dataframe cache"
        )


@typechecked
def _eval_Construct(instruction: Construct) -> DataFrame:
    return DataFrame(instruction.data)


@typechecked
def _eval_Limit(instruction: Limit, dataframe: DataFrame) -> DataFrame:
    return dataframe.head(instruction.num)


@typechecked
def _eval_Information(instruction: Information, dataframe: DataFrame) -> DataFrame:
    return variable_attributes_to_dataframe(dataframe)


@typechecked
def _eval_ProjectAttrs(instruction: ProjectAttrs, dataframe: DataFrame) -> DataFrame:
    cols = set(list(dataframe))
    invalid_attrs = set(instruction.attrs) - cols
    if invalid_attrs:
        raise InvalidAttributes(list(invalid_attrs))
    return dataframe[list(instruction.attrs)]


@typechecked
def _eval_ProjectEntity(instruction: ProjectEntity, dataframe: DataFrame) -> DataFrame:
    if instruction.ocsf_field == "event":
        df = dataframe.drop_duplicates()
    else:
        # No translation/mapping, assuming the data is already in OCSF (Kestrel extension)
        df = dataframe[
            [col for col in dataframe if col.startswith(instruction.ocsf_field)]
        ]
        df = df.rename(columns=lambda x: x[len(instruction.ocsf_field) + 1 :])
        df = df.drop_duplicates()
    return df


@typechecked
def _eval_Filter(instruction: Filter, dataframe: DataFrame) -> DataFrame:
    return dataframe[_eval_Filter_exp(instruction.exp, dataframe)]


@typechecked
def _eval_Filter_exp(exp: FExpression, dataframe: DataFrame) -> Series:
    # return: a series of boolean, same length as dataframe
    if isinstance(exp, AbsoluteTrue):
        bs = Series(True, index=dataframe.index)
    elif isinstance(exp, BoolExp):
        bs = _eval_Filter_exp_BoolExp(exp, dataframe)
    elif isinstance(exp, MultiComp):
        bss = [xs for xs in _eval_Filter_exp(exp.comps, dataframe)]
        if exp.op == ExpOp.AND:
            bs = functools.reduce(lambda x, y: x & y, bss)
        elif exp.op == ExpOp.OR:
            bs = functools.reduce(lambda x, y: x | y, bss)
        else:
            raise NotImplementedError("unkown kestrel.ir.filter.ExpOp type")
    else:
        bs = _eval_Filter_exp_Comparison(exp, dataframe)
    return bs


@typechecked
def _eval_Filter_exp_BoolExp(boolexp: BoolExp, dataframe: DataFrame) -> Series:
    # return: a series of boolean, same length as dataframe
    if boolexp.op == ExpOp.AND:
        bs = _eval_Filter_exp(boolexp.lhs, dataframe) & _eval_Filter_exp(
            boolexp.rhs, dataframe
        )
    elif boolexp.op == ExpOp.OR:
        bs = _eval_Filter_exp(boolexp.lhs, dataframe) | _eval_Filter_exp(
            boolexp.rhs, dataframe
        )
    else:
        raise NotImplementedError("unkown kestrel.ir.filter.ExpOp type")
    return bs


@typechecked
def _eval_Filter_exp_Comparison(
    c: FBasicComparison,
    df: DataFrame,
) -> Series:
    # return: a series of boolean, same length as dataframe
    comp2func = {
        NumCompOp.EQ: operator.eq,
        NumCompOp.NEQ: operator.ne,
        NumCompOp.LT: operator.gt,  # value first in functools.partial
        NumCompOp.LE: operator.ge,  # value first in functools.partial
        NumCompOp.GT: operator.lt,  # value first in functools.partial
        NumCompOp.GE: operator.le,  # value first in functools.partial
        StrCompOp.EQ: operator.eq,
        StrCompOp.NEQ: operator.ne,
        StrCompOp.LIKE: lambda w, x: bool(
            re.search(w.replace(".", r"\.").replace("%", ".*?"), x)
        ),
        StrCompOp.NLIKE: lambda w, x: not bool(
            re.search(w.replace(".", r"\.").replace("%", ".*?"), x)
        ),
        StrCompOp.MATCHES: lambda w, x: bool(re.search(w, x)),
        StrCompOp.NMATCHES: lambda w, x: not bool(re.search(w, x)),
        ListOp.IN: lambda w, x: x in w,
        ListOp.NIN: lambda w, x: x not in w,
    }

    # if c.value is from previous subquery evaluation,
    # turn it into Union[List[str], List[int], List[Tuple]]
    # TODO: may upgrade from List to Set for faster IN test
    if isinstance(c.value, DataFrame):
        if len(c.value.columns) == 1:
            c.value = list(c.value.iloc[:, 0])
        else:
            c.value = list(c.value.itertuples(index=False, name=None))

    try:
        # RefComparison has .fields; others have .field
        if isinstance(c, RefComparison):
            if len(c.fields) == 1:
                bools = df[c.fields[0]].apply(
                    functools.partial(comp2func[c.op], c.value)
                )
            else:
                if not (
                    isinstance(c.value, list)
                    and isinstance(c.value[0], tuple)
                    and len(c.fields) == len(c.value[0])
                ):
                    raise MismatchedFieldValueInMultiColumnComparison(c)

                # only support ListOp.IN and ListOp.NIN
                if c.op not in (ListOp.IN, ListOp.NIN):
                    raise InvalidOperatorInMultiColumnComparison(c)

                bools = df.set_index(c.fields).index.isin(c.value)
                # keep type consistent: from ndarray to Series
                # flip boolean if the operator is "not in"
                bools = Series(bools) if c.op == ListOp.IN else ~Series(bools)
        else:
            bools = df[c.field].apply(functools.partial(comp2func[c.op], c.value))
        return bools
    except KeyError as e:
        raise e
        raise NotImplementedError(f"unkown kestrel.ir.filter.*Op type: {c.op}")



import os
from collections import Counter
from datetime import datetime, timedelta, timezone

import pytest
from pandas import DataFrame, read_csv

from kestrel.config import load_kestrel_config
from kestrel.frontend.parser import parse_kestrel_and_update_irgraph
from kestrel.ir.filter import AbsoluteTrue, RefComparison, ReferenceValue
from kestrel.ir.graph import IRGraph
from kestrel.ir.instructions import (Construct, DataSource, Explain, Filter,
                                     Limit, Offset, ProjectAttrs,
                                     ProjectEntity, Reference, Return,
                                     SerializableDataFrame, Sort, Variable)


@pytest.fixture
def process_creation_events():
    # return a two-node graph:
    #   - Construct: table from logs_ocsf_process_creation.csv
    #   - Variable es: events pointing to the Construct
    graph = IRGraph()
    parse_kestrel_and_update_irgraph("es = NEW event [ {'id': 1} ]", graph, {})
    data_node = graph.get_nodes_by_type(Construct)[0]
    test_dir = os.path.dirname(os.path.abspath(__file__))
    data_node.data = SerializableDataFrame(read_csv(os.path.join(test_dir, "logs_ocsf_process_creation.csv")))
    return graph


@pytest.fixture
def kestrel_config():
    return load_kestrel_config()


@pytest.mark.parametrize(
    "stmt", [
        "x = GET thing FROM if://ds WHERE foo = 'bar'",
        "x = GET thing FROM if://ds WHERE foo > 1.5",
        r"x = GET thing FROM if://ds WHERE foo = r'C:\TMP'",
        "x = GET thing FROM if://ds WHERE foo = 'bar' OR baz != 42",
        "x = GET thing FROM if://ds WHERE foo = 'bar' AND baz IN (1, 2, 3)",
        "x = GET thing FROM if://ds WHERE foo = 'bar' AND baz IN (1)",
        "x = GET thing FROM if://ds WHERE foo = 'bar' AND baz IN (1) LAST 3 DAYS",
    ]
)
def test_parser_get_statements(stmt):
    """
    This test isn't meant to be comprehensive, but checks basic transformer functionality.

    This will need to be updated as we build out the new Transformer
    """

    graph = IRGraph()
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    assert len(graph) == 4
    assert len(graph.get_nodes_by_type(Variable)) == 1
    assert len(graph.get_nodes_by_type(ProjectEntity)) == 1
    assert len(graph.get_nodes_by_type(DataSource)) == 1
    assert len(graph.get_nodes_by_type(Filter)) == 1

    # Ensure result is serializable
    _ = graph.to_json()


def test_parser_get_timespan_relative():
    stmt = "x = GET url FROM if://ds WHERE url = 'http://example.com/' LAST 5h"
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    filt_list = graph.get_nodes_by_type(Filter)
    assert len(filt_list) == 1
    filt = filt_list[0]
    delta = filt.timerange.stop - filt.timerange.start
    assert delta == timedelta(hours=5)


def test_parser_get_timespan_absolute():
    stmt = ("x = GET url FROM if://ds WHERE url = 'http://example.com/'"
            " START '2023-11-29T00:00:00Z' STOP '2023-11-29T05:00:00Z'")
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    filt_list = graph.get_nodes_by_type(Filter)
    assert len(filt_list) == 1
    filt = filt_list[0]
    delta = filt.timerange.stop - filt.timerange.start
    assert delta == timedelta(hours=5)
    assert filt.timerange.start == datetime(2023, 11, 29, 0, 0, tzinfo=timezone.utc)
    assert filt.timerange.stop == datetime(2023, 11, 29, 5, 0, tzinfo=timezone.utc)


@pytest.mark.parametrize(
    "stmt, expected", [
        ("x = GET url FROM if://ds WHERE url = 'http://example.com/' LIMIT 1", 1),
        ("x = GET url FROM if://ds WHERE url = 'http://example.com/' LAST 3d LIMIT 2", 2),
        (("x = GET url FROM if://ds WHERE url = 'http://example.com/'"
          " START '2023-11-29T00:00:00Z' STOP '2023-11-29T05:00:00Z' LIMIT 3"), 3),
    ]
)
def test_parser_get_with_limit(stmt, expected):
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    limits = graph.get_nodes_by_type(Limit)
    assert len(limits) == 1
    limit = limits[0]
    assert limit.num == expected


def get_parsed_filter_exp(stmt):
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    filter_node = graph.get_nodes_by_type(Filter).pop()
    return filter_node.exp


def test_parser_mapping_single_comparison():
    # test when entity name is not included in the attributes
    stmt = "x = GET process FROM if://ds WHERE binary_ref.name = 'foo'"
    parse_filter = get_parsed_filter_exp(stmt)
    assert parse_filter.field == 'process.file.name'
    # another test
    stmt = "x = GET ipv4-addr FROM if://ds WHERE value = '192.168.22.3'"
    parse_filter = get_parsed_filter_exp(stmt)
    assert parse_filter.field == 'device.ip'

    # this is a special case in parser logic
    # if the field already has entity prefix, do not filter it with entity type
    # since extended graph (from non-return entity) can be put here
    stmt = "x = GET process FROM if://ds WHERE process:binary_ref.name = 'foo'"
    fields = set([x.field for x in get_parsed_filter_exp(stmt).comps])
    assert fields == set(('process.file.name', 'actor.process.file.name'))


def test_parser_stix_mapping_network_traffic():
    stmt = "x = GET network-traffic FROM if://ds WHERE src_ref.value = '192.168.22.3'"
    parse_filter = get_parsed_filter_exp(stmt)
    assert parse_filter.field == 'src_endpoint.ip'


def test_parser_mapping_multiple_comparison_to_multiple_values():
    stmt = "x = GET process FROM if://ds WHERE binary_ref.name = 'foo' "\
        "OR name = 'bam' AND parent_ref.name = 'boom'"
    parse_filter = get_parsed_filter_exp(stmt)
    field1 = parse_filter.lhs.field
    assert field1 == 'process.file.name'
    field2 = parse_filter.rhs.lhs.field
    assert field2 == 'process.name'  # 'process.name'
    field3 = parse_filter.rhs.rhs.field
    assert field3 == "process.parent_process.name"


def test_parser_new_json():
    stmt = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
"""
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    cs = graph.get_nodes_by_type(Construct)
    assert len(cs) == 1
    construct = cs[0]
    df = DataFrame([ {"name": "cmd.exe", "pid": 123}
         , {"name": "explorer.exe", "pid": 99}
         , {"name": "firefox.exe", "pid": 201}
         , {"name": "chrome.exe", "pid": 205}
         ])
    assert df.equals(construct.data)
    vs = graph.get_variables()
    assert len(vs) == 1
    assert vs[0].name == "proclist"


@pytest.mark.parametrize(
    "stmt, node_cnt", [
        ("x = y WHERE foo = 'bar'", 4),
        ("x = y WHERE foo > 1.5", 4),
        (r"x = y WHERE foo = r'C:\TMP'", 4),
        ("x = y WHERE foo = 'bar' OR baz != 42", 4),
        ("x = y WHERE foo = 'bar' AND baz IN (1, 2, 3)", 4),
        ("x = y WHERE foo = 'bar' AND baz IN (1)", 4),
        ("x = y WHERE foo = 'bar' SORT BY foo ASC LIMIT 3", 6),
        ("x = y WHERE foo = 'bar' SORT BY foo ASC LIMIT 3 OFFSET 9", 7),
    ]
)
def test_parser_expression(stmt, node_cnt):
    """
    This test isn't meant to be comprehensive, but checks basic transformer functionality.

    This will need to be updated as we build out the new Transformer
    """

    graph = IRGraph()
    parse_kestrel_and_update_irgraph('y = NEW process [ {"asdf": "abc.exe"} ]', graph, {})
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    assert len(graph) == node_cnt
    assert len(graph.get_nodes_by_type(Variable)) == 2
    assert len(graph.get_nodes_by_type(Filter)) == 1
    assert len(graph.get_nodes_by_type(Sort)) in (0, 1)
    assert len(graph.get_nodes_by_type(Limit)) in (0, 1)
    assert len(graph.get_nodes_by_type(Offset)) in (0, 1)


def test_three_statements_in_a_line():
    stmt = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
browsers = proclist WHERE name = 'firefox.exe' OR name = 'chrome.exe'
DISP browsers ATTR name, pid
"""
    graph = IRGraph()
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    assert len(graph) == 6
    c = graph.get_nodes_by_type(Construct)[0]
    assert {"proclist", "browsers"} == {v.name for v in graph.get_variables()}
    proclist = graph.get_variable("proclist")
    browsers = graph.get_variable("browsers")
    proj = graph.get_nodes_by_type(ProjectAttrs)[0]
    assert proj.attrs == ('name', 'pid')
    ft = graph.get_nodes_by_type(Filter)[0]
    assert ft.exp.to_dict() == {"lhs": {"field": "name", "op": "=", "value": "firefox.exe"}, "op": "OR", "rhs": {"field": "name", "op": "=", "value": "chrome.exe"}}
    assert len(graph.edges) == 5
    assert (c, proclist) in graph.edges
    assert (proclist, ft) in graph.edges
    assert (ft, browsers) in graph.edges
    assert (browsers, proj) in graph.edges
    assert (proj, rets[0]) in graph.edges


@pytest.mark.parametrize(
    "stmt, node_cnt, expected", [
        ("x = y WHERE foo = y.foo", 5, [ReferenceValue("y", ("foo",))]),
        ("x = y WHERE foo > 1.5", 4, []),
        ("x = y WHERE foo = 'bar' OR baz = y.baz", 5, [ReferenceValue("y", ("baz",))]),
        ("x = y WHERE (foo = 'bar' OR baz = y.baz) AND (fox = y.fox AND bbb = y.bbb)", 7, [ReferenceValue("y", ("baz",)), ReferenceValue("y", ("fox",)), ReferenceValue("y", ("bbb",))]),
        ("x = GET process FROM s://x WHERE foo = y.foo", 7, [ReferenceValue("y", ("foo",))]),
        ("x = GET file FROM s://y WHERE foo > 1.5", 6, []),
        ("x = GET file FROM c://x WHERE foo = 'bar' OR baz = y.baz", 7, [ReferenceValue("y", ("baz",))]),
        ("x = GET user FROM s://x WHERE (foo = 'bar' OR baz = y.baz) AND (fox = y.fox AND bbb = y.bbb)", 9, [ReferenceValue("y", ("baz",)), ReferenceValue("y", ("fox",)), ReferenceValue("y", ("bbb",))]),
    ]
)
def test_reference_branch(stmt, node_cnt, expected):
    graph = IRGraph()
    parse_kestrel_and_update_irgraph('y = NEW process [ {"asdf": "abc.exe"} ]', graph, {})
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    assert len(graph) == node_cnt
    filter_nodes = graph.get_nodes_by_type(Filter)
    assert len(filter_nodes) == 1
    filter_node = filter_nodes[0]
    for rv in expected:
        r = graph.get_variable(rv.reference)
        assert r
        projs = [p for p in graph.successors(r) if isinstance(p, ProjectAttrs) and p.attrs == rv.attributes]
        assert projs and len(projs) == 1
        proj = projs[0]
        assert proj
        assert list(graph.successors(proj)) == [filter_node]


def test_parser_disp_after_new():
    stmt = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
DISP proclist ATTR name, pid LIMIT 2 OFFSET 3
"""
    graph = IRGraph()
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    assert len(graph) == 6
    c = graph.get_nodes_by_type(Construct)[0]
    assert {"proclist"} == {v.name for v in graph.get_variables()}
    proclist = graph.get_variable("proclist")
    proj = graph.get_nodes_by_type(ProjectAttrs)[0]
    assert proj.attrs == ('name', 'pid')
    limit = graph.get_nodes_by_type(Limit)[0]
    assert limit.num == 2
    offset = graph.get_nodes_by_type(Offset)[0]
    assert offset.num == 3
    ret = rets[0]
    assert len(graph.edges) == 5
    assert (c, proclist) in graph.edges
    assert (proclist, proj) in graph.edges
    assert (proj, limit) in graph.edges
    assert (limit, offset) in graph.edges
    assert (offset, ret) in graph.edges


def test_parser_explain_alone():
    stmt = 'y = NEW process [ {"asdf": "abc.exe"} ] EXPLAIN y'
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    assert len(graph) == 4
    assert len(graph.edges) == 3
    assert Counter(map(type, graph.nodes())) == Counter([Construct, Variable, Explain, Return])


def test_parser_explain_dereferred():
    stmt = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
EXPLAIN proclist
"""
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    assert len(graph) == 4
    assert len(graph.edges) == 3
    assert Counter(map(type, graph.nodes())) == Counter([Construct, Variable, Explain, Return])


@pytest.mark.parametrize(
    "stmt, entity, ocsf_proj_field, key", [
        ("x = GET registry FROM if://ds WHERE key = 'bar'", "reg_key", "reg_key", "reg_key.path"),
        ("x = GET user-account FROM if://ds WHERE user_id = '123'", "user", "user", "user.uid"),
        ("x = GET host FROM if://ds WHERE name = 'bar'", "endpoint", "device", "device.name"),
        ("x = GET destination FROM if://ds WHERE mac = '22:33:44:55'", "network_endpoint", "dst_endpoint", "dst_endpoint.mac"),
        ("x = GET registry FROM if://ds WHERE key = 'bar'", "reg_key", "reg_key", "reg_key.path"),
        ("x = GET process.parent.user FROM if://ds WHERE name = 'alice'", "user", "process.parent_process.user", "process.parent_process.user.name"),
        ("x = GET process:parent_ref FROM if://ds WHERE command_line = 'cmd.exe -abc'", "process", "process.parent_process", "process.parent_process.cmd_line"),
    ]
)
def test_parser_entity_and_proj_and_field_mapping(stmt, entity, ocsf_proj_field, key):
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    assert graph.get_nodes_by_type(Variable)[0].entity_type == entity
    assert graph.get_nodes_by_type(ProjectEntity)[0].ocsf_field == ocsf_proj_field
    assert graph.get_nodes_by_type(Filter)[0].exp.field == key


def test_parser_find_event_to_entity(process_creation_events):
    graph = process_creation_events
    stmt = "procs = FIND process RESPONDED es"
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    assert Counter(map(type, graph.nodes())) == Counter([Construct, Variable, Filter, ProjectEntity, Variable])
    filt = graph.get_nodes_by_type(Filter)[0]
    assert isinstance(filt.exp, AbsoluteTrue)
    projent = graph.get_nodes_by_type(ProjectEntity)[0]
    assert projent.ocsf_field == projent.native_field == "process"


def test_parser_find_entity_to_event(process_creation_events, kestrel_config):
    graph = process_creation_events
    stmt = """
        procs = FIND process RESPONDED es
        eves = FIND event ORIGINATED BY procs
    """
    parse_kestrel_and_update_irgraph(stmt, graph, kestrel_config["entity_identifier"])
    assert Counter(map(type, graph.nodes())) == Counter([Construct, Variable, Filter, ProjectEntity, Variable, ProjectAttrs, Filter, Variable, ProjectEntity])
    projattr = graph.get_nodes_by_type(ProjectAttrs)[0]
    filt = [f for f in graph.get_nodes_by_type(Filter) if not isinstance(f.exp, AbsoluteTrue)][0]
    assert (projattr, filt) in graph.edges
    exp = filt.exp
    assert isinstance(exp, RefComparison)
    assert set(exp.fields) == set(['actor.process.uid', 'actor.process.endpoint.uid'])
    assert exp.value.reference == 'procs'
    assert set(exp.value.attributes) == set(['uid', 'endpoint.uid'])


def test_parser_find_entity_to_entity(process_creation_events, kestrel_config):
    graph = process_creation_events
    stmt = """
        procs = FIND process RESPONDED es
        parents = FIND process CREATED procs
    """
    parse_kestrel_and_update_irgraph(stmt, graph, kestrel_config["entity_identifier"])
    assert Counter(map(type, graph.nodes())) == Counter([Construct, Variable, Filter, ProjectEntity, Variable, ProjectAttrs, Filter, ProjectEntity, Variable])
    projattr = graph.get_nodes_by_type(ProjectAttrs)[0]
    filt = [f for f in graph.get_nodes_by_type(Filter) if not isinstance(f.exp, AbsoluteTrue)][0]
    assert (projattr, filt) in graph.edges
    exp = filt.exp
    assert isinstance(exp, RefComparison)
    assert set(exp.fields) == set(['process.uid', 'process.endpoint.uid'])
    assert exp.value.reference == 'procs'
    assert set(exp.value.attributes) == set(['uid', 'endpoint.uid'])
    parents = graph.get_variable("parents")
    projent = list(graph.predecessors(parents))[0]
    assert projent.ocsf_field == "process.parent_process"
    assert projent.native_field == "process.parent_process"
